from .examples.tasks.sumlist import SumList
from ewokscore.progress import TextProgress
from ewokscore.events import job_context
from ewokscore.events import workflow_context
import numpy.random


def test_no_progress_stdout(capsys):
    task = SumList(
        inputs={"list": numpy.random.random(10000)},
    )
    assert not task.done
    task.execute()
    stdout = capsys.readouterr()
    assert len(stdout.out) == 0
    assert task.done


def test_text_progress_stdout(capsys):
    task = SumList(
        inputs={"list": numpy.random.random(10000)},
        progress=TextProgress(name="SumList"),
    )
    assert not task.done
    task.execute()
    stdout = capsys.readouterr()
    assert len(stdout.out) > 0
    assert stdout.out.count("DONE") > 0
    assert task.done


def test_no_progress_events(caplog):
    with job_context(None) as execinfo:
        with workflow_context(execinfo) as execinfo:
            task = SumList(
                inputs={"list": numpy.random.random(10000)},
                node_id="node_id",
                execinfo=execinfo,
            )
            task.execute()
    assert len(caplog.records) == 6


def test_text_progress_events(caplog):
    with job_context(None) as execinfo:
        with workflow_context(execinfo) as execinfo:
            task = SumList(
                inputs={"list": numpy.random.random(10000)},
                progress=TextProgress(name="SumList"),
                node_id="node_id",
                execinfo=execinfo,
            )
            task.execute()
    nprogress = 100
    assert len(caplog.records) == (nprogress + 6)
    progress = [
        record.progress for record in caplog.records if record.type == "progress"
    ]
    assert len(progress) == nprogress
    assert progress == list(range(1, nprogress + 1))



import os
import sys
import pytest
from ewokscore.task import Task
from .examples import tasks


@pytest.fixture(scope="module")
def notebook():
    return os.path.join(tasks.__path__[0], "notebooktask.ipynb")


@pytest.mark.xfail(sys.version_info < (3, 8), reason="papermill #711", strict=False)
def test_notebook_task(notebook, testkernel):
    task = Task.instantiate(
        "NotebookExecutorTask",
        inputs={
            "_notebook": notebook,
            "_kernel_name": testkernel,
            "_execution_timeout": 60,
            "a": 3,
            "b": 5,
        },
    )
    task.execute()
    assert task.done
    assert task.succeeded
    expected = {
        "output_notebook": None,
        "results": {"result": 8, "x": "string", "y": 10, "z": [1, 2, 3]},
    }
    assert task.get_output_values() == expected


@pytest.mark.xfail(sys.version_info < (3, 8), reason="papermill #711", strict=False)
def test_notebook_task_save(tmpdir, notebook, testkernel):
    output_notebook = tmpdir / "nb.ipynb"
    notebook = os.path.join(tasks.__path__[0], "notebooktask.ipynb")
    task = Task.instantiate(
        "NotebookExecutorTask",
        inputs={
            "_notebook": notebook,
            "_output_notebook": str(output_notebook),
            "_kernel_name": testkernel,
            "_execution_timeout": 60,
            "a": 3,
            "b": 5,
        },
    )
    task.execute()
    assert task.done
    assert task.succeeded
    expected = {
        "output_notebook": str(output_notebook),
        "results": {"result": 8, "x": "string", "y": 10, "z": [1, 2, 3]},
    }
    assert task.get_output_values() == expected
    assert output_notebook.exists()


@pytest.mark.xfail(sys.version_info < (3, 8), reason="papermill #711", strict=False)
def test_notebook_task_failure(notebook, testkernel):
    task = Task.instantiate(
        "NotebookExecutorTask",
        inputs={
            "_notebook": notebook,
            "_kernel_name": testkernel,
            "_execution_timeout": 60,
            "a": "wrong value",
            "b": 10,
        },
    )
    with pytest.raises(RuntimeError):
        task.execute()
    assert task.done
    assert task.failed



from contextlib import contextmanager
from logging.handlers import QueueHandler
from queue import Queue, Empty
import pytest
from ewokscore import events


@contextmanager
def capture_events(blocking):
    queue = Queue()
    handler = QueueHandler(queue)
    events.add_handler(handler, blocking)

    def get_event():
        try:
            return queue.get(block=blocking, timeout=1)
        except Empty:
            raise RuntimeError("event not received by handler") from None

    try:
        yield get_event
    finally:
        events.remove_handler(handler)
        events.cleanup()


@pytest.mark.parametrize("blocking", [False, True])
def test_workflow_event(blocking):
    execinfo = {
        "job_id": None,
        "host_name": None,
        "user_name": None,
        "process_id": None,
        "workflow_id": None,
    }
    with capture_events(blocking) as get_event:
        events.send_workflow_event(execinfo=execinfo, event="start")
        event = get_event()
        assert event.type == "start"

        events.send_workflow_event(execinfo=execinfo, event="end", error_message="abc")
        event = get_event()
        assert event.type == "end"
        assert event.error
        assert event.error_message == "abc"


@pytest.mark.parametrize("blocking", [False, True])
def test_task_event(blocking):
    execinfo = {
        "job_id": None,
        "host_name": None,
        "user_name": None,
        "process_id": None,
        "workflow_id": None,
        "node_id": None,
        "task_id": None,
    }
    with capture_events(blocking) as get_event:
        events.send_task_event(
            execinfo=execinfo,
            event="start",
        )
        event = get_event()
        assert event.type == "start"

        events.send_task_event(
            execinfo=execinfo,
            event="progress",
            progress=50,
        )
        event = get_event()
        assert event.type == "progress"
        assert event.progress == 50

        events.send_task_event(
            execinfo=execinfo,
            event="end",
        )
        event = get_event()
        assert event.type == "end"
        assert not event.error
        assert event.error_message is None



from ewokscore import Task


class MyTask1(
    Task, input_names=["a"], optional_input_names=["b"], output_names=["result"]
):
    """Test 1"""

    def run(self):
        pass


class MyTask2(
    Task, input_names=["a"], optional_input_names=["b"], output_names=["result"]
):
    def run(self):
        pass


def run(a, b=None):
    """Test 2"""
    pass


def myfunc(a, b=None):
    pass


def _myfunc(a, b=None):
    pass



from typing import Iterable, Optional, Tuple
import pytest

from ewokscore import execute_graph
from ewokscore import load_graph
from ewokscore import convert_graph
from ewokscore import graph_is_supported
from ewokscore.graph.analysis import start_nodes

from .examples.graphs import graph_names
from .examples.graphs import get_graph
from .utils.results import assert_execute_graph_default_result
from .utils.results import assert_execute_graph_tasks
from .utils.show import show_graph


@pytest.mark.parametrize("graph_name", graph_names())
@pytest.mark.parametrize("scheme", (None, "json", "nexus"))
def test_execute_graph(graph_name, scheme, tmpdir):
    graph, expected = get_graph(graph_name)
    ewoksgraph = load_graph(graph)
    if scheme:
        varinfo = {"root_uri": str(tmpdir), "scheme": scheme}
    else:
        varinfo = None
    if not graph_is_supported(ewoksgraph):
        with pytest.raises(RuntimeError):
            execute_graph(ewoksgraph, varinfo=varinfo)
        return

    result = execute_graph(ewoksgraph, varinfo=varinfo)
    assert_execute_graph_default_result(ewoksgraph, result, expected, varinfo)
    result = execute_graph(ewoksgraph, varinfo=varinfo, output_tasks=True)
    assert_execute_graph_tasks(ewoksgraph, result, expected, varinfo)


def test_graph_cyclic():
    graph, _ = get_graph("empty")
    ewoksgraph = load_graph(graph)
    assert not ewoksgraph.is_cyclic
    graph, _ = get_graph("acyclic1")
    ewoksgraph = load_graph(graph)
    assert not ewoksgraph.is_cyclic
    graph, _ = get_graph("cyclic1")
    ewoksgraph = load_graph(graph)
    assert ewoksgraph.is_cyclic


def test_start_nodes():
    graph, _ = get_graph("acyclic1")
    ewoksgraph = load_graph(graph)
    assert start_nodes(ewoksgraph.graph) == {"task1", "task2"}

    graph, _ = get_graph("acyclic2")
    ewoksgraph = load_graph(graph)
    assert start_nodes(ewoksgraph.graph) == {"task1"}

    graph, _ = get_graph("cyclic1")
    ewoksgraph = load_graph(graph)
    assert start_nodes(ewoksgraph.graph) == {"task1"}

    graph, _ = get_graph("triangle1")
    ewoksgraph = load_graph(graph)
    assert start_nodes(ewoksgraph.graph) == {"task1"}


@pytest.mark.parametrize("graph_name", graph_names())
@pytest.mark.parametrize(
    "representation", (None, "json", "json_dict", "json_string", "yaml")
)
def test_serialize_graph(graph_name, representation, tmpdir):
    graph, _ = get_graph(graph_name)
    ewoksgraph = load_graph(graph)
    if representation == "yaml":
        destination = str(tmpdir / "file.yml")
    elif representation == "json":
        destination = str(tmpdir / "file.json")
    else:
        destination = None
    inmemorydump = ewoksgraph.dump(destination, representation=representation)

    if destination:
        source = destination
    else:
        source = inmemorydump
    ewoksgraph2 = load_graph(source, representation=representation)

    assert ewoksgraph == ewoksgraph2


@pytest.mark.parametrize("graph_name", graph_names())
def test_convert_graph(graph_name, tmpdir):
    graph, _ = get_graph(graph_name)
    ewoksgraph = load_graph(graph)
    assert_convert_graph(convert_graph, ewoksgraph, tmpdir)


def assert_convert_graph(
    convert_graph,
    ewoksgraph,
    tmpdir,
    representations: Optional[Iterable[Tuple[dict, dict, Optional[str]]]] = None,
):
    """All graph `representations` need to be known by `convert_graph`. It will always
    test the basic representations (e.g. json and yaml) in addition to the provided
    `representations`.

    The tuple-items in `representations` are: load options, save options, file extension.
    """
    non_serialized_representation = dict(), dict(), None
    conversion_chain = [
        non_serialized_representation,
        (dict(), {"representation": "json"}, "json"),
        (dict(), {"representation": "yaml"}, "yaml"),
        (dict(), {"representation": "json_dict"}, None),
        (dict(), {"representation": "json_string"}, None),
    ]
    if representations:
        conversion_chain.extend(representations)
    conversion_chain.append(non_serialized_representation)
    source = ewoksgraph
    for convert_from, convert_to in zip(conversion_chain[:-1], conversion_chain[1:]):
        load_options, _, _ = convert_from
        _, save_options, fileext = convert_to
        if fileext:
            destination = str(tmpdir / f"file.{fileext}")
        else:
            destination = None
        result = convert_graph(
            source,
            destination,
            load_options=load_options,
            save_options=save_options,
        )
        if fileext:
            source = destination
        else:
            source = result

    ewoksgraph2 = load_graph(source)
    try:
        assert ewoksgraph == ewoksgraph2
    except AssertionError:
        show_graph(ewoksgraph, plot=False)
        show_graph(ewoksgraph2, plot=False)
        raise



from . import graph


@graph
def demo():
    sumtask = "ewokscore.tests.examples.tasks.sumtask.SumTask"
    sumlist = "ewokscore.tests.examples.tasks.sumlist.SumList"
    nodes = [
        {
            "id": "task0",
            "default_inputs": [
                {"name": "list", "value": [0, 1, 2]},
                {"name": "delay", "value": 0},
            ],
            "task_type": "class",
            "task_identifier": sumlist,
            "ows": {"position": "(200.0, 433.0)"},
            "uiProps": {"position": {"x": 200, "y": 433}},
        },
        {
            "id": "task1",
            "default_inputs": [
                {"name": "a", "value": 1},
                {"name": "delay", "value": 0},
            ],
            "task_type": "class",
            "task_identifier": sumtask,
            "ows": {"position": "(456.0, 431.0)"},
            "uiProps": {"position": {"x": 456, "y": 431}},
        },
        {
            "id": "task2",
            "default_inputs": [
                {"name": "a", "value": 2},
                {"name": "delay", "value": 0},
            ],
            "task_type": "class",
            "task_identifier": sumtask,
            "ows": {"position": "(450.0, 155.0)"},
            "uiProps": {"position": {"x": 450, "y": 155}},
        },
        {
            "id": "task3",
            "default_inputs": [
                {"name": "b", "value": 3},
                {"name": "delay", "value": 0},
            ],
            "task_type": "class",
            "task_identifier": sumtask,
            "ows": {"position": "(709.0, 432.0)"},
            "uiProps": {"position": {"x": 709, "y": 432}},
        },
        {
            "id": "task4",
            "default_inputs": [
                {"name": "b", "value": 4},
                {"name": "delay", "value": 0},
            ],
            "task_type": "class",
            "task_identifier": sumtask,
            "ows": {"position": "(705.0, 152.0)"},
            "uiProps": {"position": {"x": 705, "y": 152}},
        },
        {
            "id": "task5",
            "default_inputs": [
                {"name": "b", "value": 5},
                {"name": "delay", "value": 0},
            ],
            "task_type": "class",
            "task_identifier": sumtask,
            "ows": {"position": "(902.0, 264.0)"},
            "uiProps": {"position": {"x": 902, "y": 264}},
        },
        {
            "id": "task6",
            "default_inputs": [
                {"name": "b", "value": 6},
                {"name": "delay", "value": 0},
            ],
            "task_type": "class",
            "task_identifier": sumtask,
            "ows": {"position": "(1113.0, 263.0)"},
            "uiProps": {"position": {"x": 1113, "y": 263}},
        },
    ]

    links = [
        {
            "source": "task0",
            "target": "task1",
            "data_mapping": [{"target_input": "a", "source_output": "sum"}],
        },
        {
            "source": "task1",
            "target": "task3",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task4",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task3",
            "target": "task5",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task4",
            "target": "task5",
            "data_mapping": [{"target_input": "b", "source_output": "result"}],
        },
        {
            "source": "task5",
            "target": "task6",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
    ]

    graph = {
        "links": links,
        "nodes": nodes,
    }

    expected_results = {
        "task0": {"sum": 3},
        "task1": {"result": 3},
        "task2": {"result": 2},
        "task3": {"result": 6},
        "task4": {"result": 6},
        "task5": {"result": 12},
        "task6": {"result": 18},
    }

    return graph, expected_results



import os
import enum
import json
import yaml
import logging
from typing import Optional, Union
from collections.abc import Mapping
import importlib

import networkx
from ewoksutils.path_utils import makedirs_from_filename

from ..node import node_id_from_json
from .schema import normalize_schema_version

logger = logging.getLogger(__name__)

GraphRepresentation = enum.Enum(
    "GraphRepresentation", "json json_dict json_string json_module yaml"
)


def _ewoks_jsonload_hook_pair(item):
    key, value = item
    if key in (
        "source",
        "target",
        "sub_source",
        "sub_target",
        "id",
        "node",
        "sub_node",
    ):
        value = node_id_from_json(value)
    return key, value


def ewoks_jsonload_hook(items):
    return dict(map(_ewoks_jsonload_hook_pair, items))


def graph_full_path(path, root_dir=None, root_module=None, possible_extensions=tuple()):
    if not root_dir and root_module:
        root_dir = _package_path(root_module)
    if not os.path.isabs(path) and root_dir:
        path = os.path.join(root_dir, path)
    path = os.path.abspath(path)
    if os.path.exists(path):
        return path
    root, _ = os.path.splitext(path)
    for new_ext in possible_extensions:
        new_full_path = root + new_ext
        if os.path.exists(new_full_path):
            return new_full_path
    raise FileNotFoundError(path)


def dump(
    graph: networkx.DiGraph,
    destination=None,
    representation: Optional[Union[GraphRepresentation, str]] = None,
    **kw,
) -> Union[str, dict]:
    """From runtime to persistent representation"""
    if isinstance(representation, str):
        representation = GraphRepresentation.__members__[representation]
    if representation is None:
        if isinstance(destination, str):
            filename = destination.lower()
            if filename.endswith(".json"):
                representation = GraphRepresentation.json
            elif filename.endswith((".yml", ".yaml")):
                representation = GraphRepresentation.yaml
        else:
            representation = GraphRepresentation.json_dict
    if representation == GraphRepresentation.json_dict:
        return _networkx_to_dict(graph)
    elif representation == GraphRepresentation.json:
        dictrepr = dump(graph)
        makedirs_from_filename(destination)
        kw.setdefault("indent", 2)
        with open(destination, mode="w") as f:
            json.dump(dictrepr, f, **kw)
        return destination
    elif representation == GraphRepresentation.json_string:
        dictrepr = dump(graph)
        return json.dumps(dictrepr, **kw)
    elif representation == GraphRepresentation.yaml:
        dictrepr = dump(graph)
        makedirs_from_filename(destination)
        with open(destination, mode="w") as f:
            yaml.dump(dictrepr, f, **kw)
        return destination
    elif representation == GraphRepresentation.json_module:
        package, _, file = destination.rpartition(".")
        assert package, f"No package provided when saving graph to '{destination}'"
        destination = os.path.join(_package_path(package), f"{file}.json")
        return dump(graph, destination=destination, representation="json", **kw)
    else:
        raise TypeError(representation, type(representation))


def load(
    source=None,
    representation: Optional[Union[GraphRepresentation, str]] = None,
    root_dir: Optional[str] = None,
    root_module: Optional[str] = None,
) -> networkx.DiGraph:
    """From persistent to runtime representation"""
    if isinstance(representation, str):
        representation = GraphRepresentation.__members__[representation]
    if representation is None:
        if isinstance(source, Mapping):
            representation = GraphRepresentation.json_dict
        elif isinstance(source, str):
            if "{" in source and "}" in source:
                representation = GraphRepresentation.json_string
            else:
                filename = source.lower()
                if filename.endswith(".json"):
                    representation = GraphRepresentation.json
                elif filename.endswith((".yml", ".yaml")):
                    representation = GraphRepresentation.yaml
                else:
                    representation = GraphRepresentation.json
    if not source:
        graph = networkx.DiGraph()
    elif isinstance(source, networkx.Graph):
        graph = source
    elif hasattr(source, "graph") and isinstance(source.graph, networkx.Graph):
        graph = source.graph
    elif representation == GraphRepresentation.json_dict:
        graph = _dict_to_networkx(source)
    elif representation == GraphRepresentation.json:
        source = graph_full_path(
            source,
            root_dir=root_dir,
            root_module=root_module,
            possible_extensions=(".json",),
        )
        with open(source, mode="r") as f:
            source = json.load(f, object_pairs_hook=ewoks_jsonload_hook)
        graph = _dict_to_networkx(source)
    elif representation == GraphRepresentation.json_string:
        source = json.loads(source, object_pairs_hook=ewoks_jsonload_hook)
        graph = _dict_to_networkx(source)
    elif representation == GraphRepresentation.yaml:
        source = graph_full_path(
            source,
            root_dir=root_dir,
            root_module=root_module,
            possible_extensions=(".yml", ".yaml"),
        )
        with open(source, mode="r") as f:
            source = yaml.load(f, yaml.Loader)
        graph = _dict_to_networkx(source)
    elif representation == GraphRepresentation.json_module:
        package, _, source = source.rpartition(".")
        if package:
            source = os.path.join(_package_path(package), source)
        return load(
            source,
            representation="json",
            root_dir=root_dir,
            root_module=root_module,
        )
    else:
        raise TypeError(representation, type(representation))

    if not networkx.is_directed(graph):
        raise TypeError(graph, type(graph))

    return graph


def _package_path(package: str) -> str:
    package = importlib.import_module(package)
    return package.__path__[0]


def _dict_to_networkx(graph: dict) -> networkx.DiGraph:
    graph.setdefault("directed", True)
    graph.setdefault("nodes", list())
    graph.setdefault("links", list())
    graph.setdefault("graph", dict())

    if "id" not in graph["graph"]:
        logger.warning('Graph has no "id": use "notspecified"')
        graph["graph"]["id"] = "notspecified"
    normalize_schema_version(graph)

    return networkx.readwrite.json_graph.node_link_graph(graph)


def _networkx_to_dict(graph: networkx.DiGraph) -> dict:
    return networkx.readwrite.json_graph.node_link_data(graph)



from typing import Dict, Iterator, List, Mapping, Optional, Union
import networkx

from .analysis import start_nodes
from .analysis import end_nodes
from ..node import NodeIdType
from ..node import get_node_label
from .. import missing_data
from ..task import Task


def update_default_inputs(
    graph: networkx.DiGraph, inputs: Optional[List[dict]] = None
) -> None:
    """Input items have the following keys:

    - name: input variable name
    - value: input variable value
    - id (optional): node id
    - label (optional): used when `id` is missing
    - task_identifier (optional): used when `id` is missing
    - all (optional): used when `id`, `label` and `task_identifier` are missing (`True`: all nodes, `False`: start nodes)
    """
    inputs = parse_inputs(graph, inputs)
    keys_to_update = "name", "value"
    for input_item in inputs:
        node_id = input_item.get("id")
        if node_id is None:
            continue
        node_attrs = graph.nodes[node_id]
        default_inputs = node_attrs.get("default_inputs")
        input_item = {k: input_item[k] for k in keys_to_update}
        if default_inputs:
            for existing_input_item in default_inputs:
                if existing_input_item["name"] == input_item["name"]:
                    existing_input_item.update(input_item)
                    break
            else:
                default_inputs.append(input_item)
        else:
            node_attrs["default_inputs"] = [input_item]


def parse_inputs(
    graph: networkx.DiGraph, inputs: Optional[List[dict]] = None
) -> List[dict]:
    """Input items have the following keys:

    - name: input variable name
    - value: input variable value
    - id (optional): node id
    - label (optional): used when `id` is missing
    - task_identifier (optional): used when `id` is missing
    - all (optional): used when `id`, `label` and `task_identifier` are missing (`True`: all nodes, `False`: start nodes)
    """
    if not inputs:
        return list()
    required = {"name", "value"}
    returned = {"id", "name", "value"}
    parsed = list()
    for input_item in list(inputs):
        missing = required - input_item.keys()
        if missing:
            raise ValueError(f"missing keys in one of the graph inputs: {missing}")
        if "id" in input_item:
            parsed.append({k: v for k, v in input_item.items() if k in returned})
            continue

        node_filters = dict()
        for k in ("label", "task_identifier"):
            if k in input_item:
                node_filters[k] = input_item[k]

        if node_filters:
            node_ids = iter_node_ids(graph, **node_filters)
        elif input_item.get("all"):
            node_ids = graph.nodes
        else:
            node_ids = start_nodes(graph)

        for node_id in node_ids:
            input_item = {k: v for k, v in input_item.items() if k in returned}
            input_item["id"] = node_id
            parsed.append(input_item)
    return parsed


def parse_outputs(
    graph: networkx.DiGraph, outputs: Optional[List[dict]] = None
) -> List[dict]:
    """Output items have the following keys:

    - name (optional): output variable name (all outputs when missing)
    - new_name (optional): optional renaming when `name` is defined
    - id (optional): node id
    - label (optional): used when `id` is missing
    - task_identifier (optional): used when `id` is missing
    - all (optional): used when `id`, `label` and `task_identifier` are missing (`True`: all nodes, `False`: end nodes)
    """
    if outputs is None:
        outputs = [{"all": False}]
    parsed = list()
    returned = {"id", "name", "new_name"}
    for output_item in outputs:
        if "id" in output_item:
            parsed.append({k: v for k, v in output_item.items() if k in returned})
            continue

        node_filters = dict()
        for k in ("label", "task_identifier"):
            if k in output_item:
                node_filters[k] = output_item[k]

        if node_filters:
            node_ids = iter_node_ids(graph, **node_filters)
        elif output_item.get("all"):
            node_ids = graph.nodes
        else:
            node_ids = end_nodes(graph)

        for node_id in node_ids:
            output_item = {k: v for k, v in output_item.items() if k in returned}
            output_item["id"] = node_id
            parsed.append(output_item)

    return parsed


def iter_node_ids(
    graph: networkx.DiGraph,
    label: Optional[str] = None,
    task_identifier: Optional[str] = None,
) -> Iterator[NodeIdType]:
    """Yield nodes with matching `label` AND `task_identifier`"""
    for node_id, node_attrs in graph.nodes.items():
        return_id = False
        if label is not None:
            node_label = get_node_label(node_id, node_attrs)
            if label != node_label:
                continue
            return_id = True
        if task_identifier is not None:
            s = node_attrs.get("task_identifier")
            if not s or not s.endswith(task_identifier):
                continue
            return_id = True
        if return_id:
            yield node_id


def extract_output_values(
    node_id: NodeIdType, task_or_outputs: Union[Task, Mapping], outputs: List[dict]
) -> Optional[dict]:
    """Output items have the following keys:

    - id: node id
    - label (optional): used when `id` is missing
    - name (optional): output variable name (all outputs when missing)
    - new_name (optional): optional renaming when name is defined
    """
    output_values = None
    if isinstance(task_or_outputs, Task):
        task_output_values = None
    else:
        task_output_values = task_or_outputs
    for output_item in outputs:
        if output_item.get("id") != node_id:
            continue
        if task_output_values is None:
            task_output_values = task_or_outputs.get_output_values()
        if output_values is None:
            output_values = dict()
        name = output_item.get("name")
        if name:
            new_name = output_item.get("new_name", name)
            output_values[new_name] = task_output_values.get(
                name, missing_data.MISSING_DATA
            )
        else:
            output_values.update(task_output_values)
    return output_values


def add_output_values(
    output_values: dict,
    node_id: NodeIdType,
    task_or_outputs: Union[Task, Dict],
    outputs: List[dict],
    merge_outputs: Optional[bool] = True,
) -> None:
    """Output items have the following keys:

    - id: node id
    - label (optional): used when `id` is missing
    - name (optional): output variable name (all outputs when missing)
    - new_name (optional): optional renaming when name is defined
    """
    task_output_values = extract_output_values(node_id, task_or_outputs, outputs)
    if task_output_values is not None:
        if merge_outputs:
            output_values.update(task_output_values)
        else:
            output_values[node_id] = task_output_values



from ewokscore import execute_graph
from ewokscore.graph import load_graph
from ewoksutils.import_utils import qualname
from ewokscore.node import node_id_as_string

from .utils.results import assert_execute_graph_tasks


def myfunc(name=None, value=0):
    print("name:", name, "value:", value)
    return value + 1


def test_sub_graph_execute():
    subsubgraph = {
        "graph": {
            "id": "subsubgraph",
            "input_nodes": [{"id": "in", "node": "subsubnode1"}],
        },
        "nodes": [
            {
                "id": "subsubnode1",
                "task_type": "method",
                "task_identifier": qualname(myfunc),
                "default_inputs": [
                    {"name": "name", "value": "subsubnode1"},
                    {"name": "value", "value": 0},
                ],
            }
        ],
    }

    subgraph = {
        "graph": {
            "id": "subgraph",
            "input_nodes": [{"id": "in", "node": "subnode1", "sub_node": "in"}],
        },
        "nodes": [
            {"id": "subnode1", "task_type": "graph", "task_identifier": subsubgraph}
        ],
    }

    graph = {
        "graph": {"id": "graph"},
        "nodes": [
            {
                "id": "node1",
                "task_type": "method",
                "task_identifier": qualname(myfunc),
                "default_inputs": [
                    {"name": "name", "value": "node1"},
                    {"name": "value", "value": 0},
                ],
            },
            {"id": "node2", "task_type": "graph", "task_identifier": subgraph},
        ],
        "links": [
            {
                "source": "node1",
                "target": "node2",
                "sub_target": "in",
                "data_mapping": [
                    {"target_input": "value", "source_output": "return_value"}
                ],
            }
        ],
    }

    ewoksgraph = load_graph(graph)
    result = execute_graph(ewoksgraph, output_tasks=True)
    expected = {
        "node1": {"return_value": 1},
        ("node2", ("subnode1", "subsubnode1")): {"return_value": 2},
    }
    assert_execute_graph_tasks(ewoksgraph, result, expected)


def test_sub_graph_link_attributes():
    subsubgraph = {
        "graph": {
            "id": "subsubgraph",
            "input_nodes": [
                {"id": "in1", "node": "subsubnode1", "link_attributes": {1: 1}},
                {"id": "in2", "node": "subsubnode1", "link_attributes": {2: 2}},
            ],
            "output_nodes": [
                {"id": "out1", "node": "subsubnode1", "link_attributes": {3: 3}},
                {"id": "out2", "node": "subsubnode1", "link_attributes": {4: 4}},
            ],
        },
        "nodes": [
            {"id": "subsubnode1", "task_type": "method", "task_identifier": "dummy"}
        ],
    }

    subgraph = {
        "graph": {
            "id": "subgraph",
            "input_nodes": [
                {
                    "id": "in1",
                    "node": "subnode1",
                    "sub_node": "in1",
                    "link_attributes": {5: 5},
                },
                {
                    "id": "in2",
                    "node": "subnode1",
                    "sub_node": "in2",
                },
            ],
            "output_nodes": [
                {
                    "id": "out1",
                    "node": "subnode1",
                    "sub_node": "out1",
                    "link_attributes": {6: 6},
                },
                {
                    "id": "out2",
                    "node": "subnode1",
                    "sub_node": "out2",
                },
            ],
        },
        "nodes": [
            {"id": "subnode1", "task_type": "graph", "task_identifier": subsubgraph}
        ],
    }

    graph = {
        "graph": {"id": "graph"},
        "nodes": [
            {"id": "node1", "task_type": "method", "task_identifier": "dummy"},
            {"id": "node2", "task_type": "method", "task_identifier": "dummy"},
            {"id": "node3", "task_type": "method", "task_identifier": "dummy"},
            {"id": "node4", "task_type": "method", "task_identifier": "dummy"},
            {"id": "graphnode", "task_type": "graph", "task_identifier": subgraph},
        ],
        "links": [
            {"source": "node1", "target": "graphnode", "sub_target": "in1"},
            {"source": "node2", "target": "graphnode", "sub_target": "in2"},
            {"source": "graphnode", "target": "node3", "sub_source": "out1"},
            {"source": "graphnode", "target": "node4", "sub_source": "out2"},
        ],
    }

    ewoksgraph = load_graph(graph)
    for link, attrs in ewoksgraph.graph.edges.items():
        numbers = {i for i in attrs if isinstance(i, int)}
        if "node1" in link:
            assert numbers == {1, 5}
        elif "node2" in link:
            assert numbers == {2}
        elif "node3" in link:
            assert numbers == {3, 6}
        elif "node4" in link:
            assert numbers == {4}
        else:
            assert False, "unexpected link"


def test_sub_graph_duplicate_aliases():
    subsubgraph = {
        "graph": {
            "id": "subsubgraph",
            "input_nodes": [
                {
                    "id": "in",
                    "node": "inode1",
                },
                {
                    "id": "in",
                    "node": "inode2",
                },
            ],
            "output_nodes": [
                {
                    "id": "out",
                    "node": "inode4",
                },
                {
                    "id": "out",
                    "node": "inode5",
                },
            ],
        },
        "nodes": [
            {"id": "inode1", "task_type": "method", "task_identifier": "dummy"},
            {"id": "inode2", "task_type": "method", "task_identifier": "dummy"},
            {"id": "inode3", "task_type": "method", "task_identifier": "dummy"},
            {"id": "inode4", "task_type": "method", "task_identifier": "dummy"},
            {"id": "inode5", "task_type": "method", "task_identifier": "dummy"},
        ],
        "links": [
            {"source": "inode1", "target": "inode3"},
            {"source": "inode2", "target": "inode3"},
            {"source": "inode3", "target": "inode4"},
            {"source": "inode3", "target": "inode5"},
        ],
    }

    subgraph = {
        "graph": {
            "id": "subgraph",
            "input_nodes": [
                {
                    "id": "in",
                    "node": "nonode",
                    "sub_node": "in",
                },
            ],
            "output_nodes": [
                {
                    "id": "out",
                    "node": "nonode",
                    "sub_node": "out",
                },
            ],
        },
        "nodes": [
            {"id": "nonode", "task_type": "graph", "task_identifier": subsubgraph}
        ],
    }

    graph = {
        "graph": {"id": "graph"},
        "nodes": [
            {"id": "node1", "task_type": "method", "task_identifier": "dummy"},
            {"id": "node2", "task_type": "method", "task_identifier": "dummy"},
            {"id": "graphnode", "task_type": "graph", "task_identifier": subgraph},
        ],
        "links": [
            {"source": "node1", "target": "graphnode", "sub_target": "in"},
            {"source": "graphnode", "target": "node2", "sub_source": "out"},
        ],
    }

    ewoksgraph = load_graph(graph)
    links = {
        (node_id_as_string(source), node_id_as_string(target))
        for source, target in ewoksgraph.graph.edges
    }
    expected = {
        ("graphnode:nonode:inode1", "graphnode:nonode:inode3"),
        ("graphnode:nonode:inode2", "graphnode:nonode:inode3"),
        ("graphnode:nonode:inode3", "graphnode:nonode:inode4"),
        ("graphnode:nonode:inode3", "graphnode:nonode:inode5"),
        ("graphnode:nonode:inode4", "node2"),
        ("graphnode:nonode:inode5", "node2"),
        ("node1", "graphnode:nonode:inode1"),
        ("node1", "graphnode:nonode:inode2"),
    }

    assert links == expected



import logging
from pprint import pformat
import sqlite3
from time import sleep
from typing import Dict, List, Optional

from ewoksutils.import_utils import qualname
from ewoksutils.sqlite3_utils import select
from ewoksutils.event_utils import FIELD_TYPES

from ewokscore import execute_graph
from ewokscore import Task
from ewokscore.events import cleanup as cleanup_events


logger = logging.getLogger(__name__)


def test_succesfull_workfow(tmpdir):
    uri = run_succesfull_workfow(tmpdir, execute_graph)
    events = fetch_events(uri, 10)
    assert_succesfull_workfow_events(events)


def test_failed_workfow(tmpdir):
    uri = run_failed_workfow(tmpdir, execute_graph)
    events = fetch_events(uri, 8)
    assert_failed_workfow_events(events)


class MyTask(
    Task, input_names=["ctr"], optional_input_names=["error_msg"], output_names=["ctr"]
):
    def run(self):
        if self.inputs.error_msg:
            raise ValueError(self.inputs.error_msg)
        else:
            self.outputs.ctr = self.inputs.ctr + 1


def run_succesfull_workfow(tmpdir, execute_graph, **execute_options):
    nodes = [
        {
            "id": "node1",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [{"name": "ctr", "value": 0}],
        },
        {
            "id": "node2",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [{"name": "ctr", "value": 0}],
        },
        {
            "id": "node3",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [{"name": "ctr", "value": 0}],
        },
    ]
    links = [
        {
            "source": "node1",
            "target": "node2",
            "data_mapping": [{"source_output": "ctr", "target_input": "ctr"}],
        },
        {
            "source": "node2",
            "target": "node3",
            "data_mapping": [{"source_output": "ctr", "target_input": "ctr"}],
        },
    ]
    graph = {"graph": {"id": "test_graph"}, "nodes": nodes, "links": links}
    return _execute_graph(tmpdir, graph, execute_graph, **execute_options)


def assert_succesfull_workfow_events(events):
    expected = [
        {"context": "job", "node_id": None, "type": "start"},
        {"context": "workflow", "node_id": None, "type": "start"},
        {"context": "node", "node_id": "node1", "type": "start"},
        {"context": "node", "node_id": "node1", "type": "end"},
        {"context": "node", "node_id": "node2", "type": "start"},
        {"context": "node", "node_id": "node2", "type": "end"},
        {"context": "node", "node_id": "node3", "type": "start"},
        {"context": "node", "node_id": "node3", "type": "end"},
        {"context": "workflow", "node_id": None, "type": "end"},
        {"context": "job", "node_id": None, "type": "end"},
    ]
    captured = [
        {k: event[k] for k in ("context", "node_id", "type")} for event in events
    ]
    _assert_events(expected, captured)


def run_failed_workfow(tmpdir, execute_graph, **execute_options):
    nodes = [
        {
            "id": "node1",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [{"name": "ctr", "value": 0}],
        },
        {
            "id": "node2",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [
                {"name": "ctr", "value": 0},
                {"name": "error_msg", "value": "abc"},
            ],
        },
        {
            "id": "node3",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [{"name": "ctr", "value": 0}],
        },
    ]
    links = [
        {
            "source": "node1",
            "target": "node2",
            "data_mapping": [{"source_output": "ctr", "target_input": "ctr"}],
        },
        {
            "source": "node2",
            "target": "node3",
            "data_mapping": [{"source_output": "ctr", "target_input": "ctr"}],
        },
    ]
    graph = {"graph": {"id": "test_graph"}, "nodes": nodes, "links": links}
    return _execute_graph(tmpdir, graph, execute_graph, **execute_options)


def assert_failed_workfow_events(events):
    expected = [
        {
            "context": "job",
            "node_id": None,
            "type": "start",
            "error_message": None,
        },
        {
            "context": "workflow",
            "node_id": None,
            "type": "start",
            "error_message": None,
        },
        {
            "context": "node",
            "node_id": "node1",
            "type": "start",
            "error_message": None,
        },
        {
            "context": "node",
            "node_id": "node1",
            "type": "end",
            "error_message": None,
        },
        {
            "context": "node",
            "node_id": "node2",
            "type": "start",
            "error_message": None,
        },
        {"context": "node", "node_id": "node2", "type": "end", "error_message": "abc"},
        {
            "context": "workflow",
            "node_id": None,
            "type": "end",
            "error_message": "Task 'node2' failed",
        },
        {
            "context": "job",
            "node_id": None,
            "type": "end",
            "error_message": "Task 'node2' failed",
        },
    ]
    captured = [
        {k: event[k] for k in ("context", "node_id", "type", "error_message")}
        for event in events
    ]
    _assert_events(expected, captured)


def _execute_graph(tmpdir, graph, execute_graph, **execute_options):
    uri = f"file:{tmpdir / 'ewoks_events.db'}"
    execinfo = execute_options.setdefault("execinfo", dict())
    handlers = execinfo.setdefault("handlers", list())
    handlers.append(
        {
            "class": "ewokscore.events.handlers.Sqlite3EwoksEventHandler",
            "arguments": [{"name": "uri", "value": uri}],
        }
    )
    cleanup_events()
    try:
        execute_graph(graph, **execute_options)
    except RuntimeError:
        pass
    return uri


def _assert_events(expected, captured):
    missing = list()
    unexpected = list(captured)
    for event in expected:
        try:
            unexpected.remove(event)
        except ValueError:
            missing.append(event)
    if missing or unexpected:
        raise AssertionError(
            f"ewoks events not as expected\nmissing:\n{pformat(missing)}\nunexpected:\n{unexpected}"
        )


def fetch_events(uri: str, nevents: int) -> List[Dict[str, Optional[str]]]:
    """Events are handled asynchronously so wait until we have the required `nevents`
    up to 3 seconds.
    """
    try:
        exception = None
        events = list()
        for _ in range(30):
            try:
                with sqlite3.connect(uri, uri=True) as conn:
                    events = list(select(conn, "ewoks_events", field_types=FIELD_TYPES))

                if len(events) != nevents:
                    raise RuntimeError(
                        f"{len(events)} ewoks events instead of {nevents}"
                    )
                return events
            except Exception as e:
                exception = e
                sleep(0.1)
        if exception:
            logger.error(exception)
        return events
    finally:
        cleanup_events()



from typing import Mapping
import networkx

from .analysis import node_pure_descendants
from ewokscore.graph import analysis


def connect_default_error_handlers(graph: networkx.DiGraph) -> networkx.DiGraph:
    """All nodes without an error handler will be connected to all default error handlers.
    Default error handlers without predecessors will be removed.
    """
    default_error_handlers = dict()
    for node_id, attrs in graph.nodes.items():
        default_error_node = attrs.pop("default_error_node", False)
        if not default_error_node:
            continue
        link_attrs = attrs.pop("default_error_attributes", None)
        if not isinstance(link_attrs, Mapping):
            link_attrs = dict()
        link_attrs["on_error"] = True
        if not (set(link_attrs.keys()) & {"map_all_data", "data_mapping"}):
            link_attrs["map_all_data"] = True
        default_error_handlers[node_id] = link_attrs

    # All nodes which are not default error handlers
    nodes_without_error_handlers = set(graph.nodes.keys()) - set(default_error_handlers)

    # Remove the nodes that already have an error handler
    for edge, attrs in graph.edges.items():
        source_id = edge[0]
        if attrs.get("on_error"):
            if source_id in nodes_without_error_handlers:
                nodes_without_error_handlers.remove(source_id)

    # Remove nodes that have any of the default error handlers as ancestor
    for node_id in set(nodes_without_error_handlers):
        for ancestor_id in analysis.node_ancestors(graph, node_id):
            if ancestor_id in default_error_handlers:
                nodes_without_error_handlers.remove(node_id)
                break

    if nodes_without_error_handlers:
        # Connect to the default error handlers
        for source_id in nodes_without_error_handlers:
            for target_id, link_attrs in default_error_handlers.items():
                graph.add_edge(source_id, target_id, **link_attrs)
    else:
        # Remove the default error handlers and their pure descendants
        for node_id in default_error_handlers:
            try:
                next(graph.predecessors())
            except StopIteration:
                nodes = set(node_pure_descendants(graph, node_id, include_node=True))
                graph.remove_nodes_from(nodes)



from ewokscore import task_discovery


def test_task_class_discovery():
    expected = [
        {
            "task_type": "class",
            "task_identifier": "ewokscore.tests.discover_module.MyTask1",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["result"],
            "category": "ewokscore",
            "description": "Test 1",
        },
        {
            "task_type": "class",
            "task_identifier": "ewokscore.tests.discover_module.MyTask2",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["result"],
            "category": "ewokscore",
            "description": None,
        },
    ]

    tasks = task_discovery.discover_tasks_from_modules()
    for task in expected:
        assert task not in tasks

    tasks = task_discovery.discover_tasks_from_modules(
        "ewokscore.tests.discover_module"
    )
    assert_tasks(tasks, expected)
    assert len(tasks) == len(expected)

    tasks = task_discovery.discover_tasks_from_modules()
    assert_tasks(tasks, expected)


def test_task_method_discovery():
    expected = [
        {
            "task_type": "method",
            "task_identifier": "ewokscore.tests.discover_module.run",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": "Test 2",
        },
        {
            "task_type": "method",
            "task_identifier": "ewokscore.tests.discover_module.myfunc",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": None,
        },
    ]
    tasks = task_discovery.discover_tasks_from_modules(
        "ewokscore.tests.discover_module", task_type="method"
    )
    assert_tasks(tasks, expected)
    assert len(tasks) == len(expected)


def test_task_ppfmethod_discovery():
    expected = [
        {
            "task_type": "method",
            "task_identifier": "ewokscore.tests.discover_module.run",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": "Test 2",
        }
    ]
    tasks = task_discovery.discover_tasks_from_modules(
        "ewokscore.tests.discover_module", task_type="ppfmethod"
    )
    assert_tasks(tasks, expected)
    assert len(tasks) == len(expected)


def test_task_discovery():
    expected = [
        {
            "category": "ewokscore",
            "optional_input_names": ["b", "delay"],
            "output_names": ["too_small", "result"],
            "required_input_names": ["a"],
            "task_identifier": "ewokscore.tests.examples.tasks.condsumtask.CondSumTask",
            "task_type": "class",
            "description": "Check whether a value is too small",
        },
        {
            "category": "ewokscore",
            "optional_input_names": ["a", "b", "raise_error"],
            "output_names": ["result"],
            "required_input_names": [],
            "task_identifier": "ewokscore.tests.examples.tasks.errorsumtask.ErrorSumTask",
            "task_type": "class",
            "description": "Add two number with intentional exception",
        },
        {
            "category": "ewokscore",
            "optional_input_names": [],
            "output_names": [],
            "required_input_names": [],
            "task_identifier": "ewokscore.tests.examples.tasks.nooutputtask.NoOutputTask",
            "task_type": "class",
            "description": "A task without outputs",
        },
        {
            "category": "ewokscore",
            "optional_input_names": ["delay"],
            "output_names": ["sum"],
            "required_input_names": ["list"],
            "task_identifier": "ewokscore.tests.examples.tasks.sumlist.SumList",
            "task_type": "class",
            "description": "Add items from a list",
        },
        {
            "category": "ewokscore",
            "optional_input_names": ["b", "delay"],
            "output_names": ["result"],
            "required_input_names": ["a"],
            "task_identifier": "ewokscore.tests.examples.tasks.sumtask.SumTask",
            "task_type": "class",
            "description": "Add two numbers with a delay",
        },
        {
            "category": "ewokscore",
            "task_identifier": "ewokscore.tests.examples.tasks.addfunc.addfunc",
            "task_type": "method",
            "required_input_names": ["arg"],
            "optional_input_names": [],
            "output_names": ["return_value"],
            "description": "Add 1 to the first argument",
        },
        {
            "category": "ewokscore",
            "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.add",
            "task_type": "method",
            "required_input_names": [],
            "optional_input_names": [],
            "output_names": ["return_value"],
            "description": "Sum objects and add 1",
        },
        {
            "category": "ewokscore",
            "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.append",
            "task_type": "method",
            "required_input_names": [],
            "optional_input_names": [],
            "output_names": ["return_value"],
            "description": "Return positional arguments as a tuple",
        },
    ]

    tasks = task_discovery.discover_all_tasks()
    assert_tasks(tasks, expected)


def assert_tasks(tasks, expected):
    listnames = ["output_names", "required_input_names", "optional_input_names"]
    for task in tasks:
        for listname in listnames:
            task[listname] = set(task[listname])

    for task in expected:
        for listname in listnames:
            task[listname] = set(task[listname])
        assert task in tasks



import os
import string
import random
from pathlib import Path
from contextlib import contextmanager
from typing import Iterable, Optional, Tuple
from silx.io import h5py_utils


def random_string(n):
    return "".join(random.choices(string.ascii_letters + string.digits, k=n))


def nonexisting_tmp_file(path: Path) -> Path:
    tmppath = path.with_name(f"tmp_ewoks_{random_string(6)}_{path.name}")
    while tmppath.exists():
        tmppath = path.with_name(f"tmp_ewoks_{random_string(6)}_{path.name}")
    return tmppath


@contextmanager
def atomic_create_path(path: Path) -> Iterable[Path]:
    """Yields a temporary path which will be renamed to the requested path
    or deleted on failure.
    """
    tmppath = nonexisting_tmp_file(path)
    tmppath.parent.mkdir(parents=True, exist_ok=True)
    try:
        yield tmppath
    except Exception:
        try:
            os.unlink(tmppath)
        except FileNotFoundError:
            pass
        raise
    while True:
        try:
            tmppath.rename(path)  # overwrite when it exists
            break
        except FileExistsError:
            path.unlink(missing_ok=True)


@contextmanager
def atomic_write(path: Path, **kw):
    with atomic_create_path(path) as tmpname:
        with open(tmpname, mode="w", **kw) as f:
            yield f


@h5py_utils.retry_contextmanager()
def append_hdf5(path: Path, **kw):
    path.parent.mkdir(parents=True, exist_ok=True)
    with h5py_utils.File(path, mode="a", **kw) as h5file:
        yield h5file


@contextmanager
def atomic_write_hdf5(
    path: Path, h5group: Optional[str], **kw
) -> Tuple[h5py_utils.File, Optional[str]]:
    if not h5group or h5group == "/":
        with atomic_create_path(path) as tmppath:
            with h5py_utils.File(tmppath, mode="a", **kw) as f:
                yield f, h5group
    else:
        with append_hdf5(path, retry_period=0.5, retry_timeout=360, **kw) as f:
            yield f, h5group



import os
import sys
import re
from pathlib import Path
from typing import Iterable, Optional, Tuple, Union
from urllib.parse import ParseResult, urlparse

WIN32 = sys.platform == "win32"


def _normalize(uri: str) -> Tuple[str, str]:
    uri = uri.replace("\\", "/")
    # Non-standard notation:
    #   "/some/path::/another/path"
    # means
    #   "/some/path?path=/another/path"
    query = ""
    query_paths = re.findall("::([^;?#]*)", uri)
    if query_paths:
        for path in query_paths:
            uri = uri.replace("::" + path, "")
        query = join_query(("path", path) for path in query_paths)
    return uri, query


def parse_uri(
    uri: str,
    default_scheme: str = "file",
    default_port: int = None,
) -> ParseResult:
    """The general structure of a URI is:
    scheme://netloc/path;parameters?query#frag
    """
    uri, query_paths = _normalize(uri)
    result = urlparse(uri)
    scheme, netloc, path, params, query, fragment = result
    if WIN32 and len(scheme) == 1:
        result = urlparse("file://" + uri)
        scheme, netloc, path, params, query, fragment = result
    query = merge_query(query_paths, query)
    if not scheme and default_scheme:
        scheme = default_scheme
    if default_port and not result.port:
        netloc = f"{result.hostname}:{default_port}"
    return type(result)(scheme, netloc, path, params, query, fragment)


def path_from_uri(uri: Union[str, ParseResult], **parse_options) -> Path:
    if isinstance(uri, str):
        uri = parse_uri(uri, **parse_options)
    return Path(uri.netloc) / uri.path


def parse_query(uri: Union[str, ParseResult], **parse_options) -> dict:
    if isinstance(uri, str):
        uri = parse_uri(uri, **parse_options)
    return split_query(uri.query)


def split_query(query: str) -> dict:
    result = dict()
    for s in query.split("&"):
        if not s:
            continue
        name, _, value = s.partition("=")
        prev_value = result.get(name)
        if prev_value:
            value = join_string(prev_value, value, "/")
        result[name] = value
    return result


def join_query(query_items: Iterable[Tuple[str, str]]) -> str:
    return "&".join(f"{k}={v}" for k, v in query_items)


def join_string(a: str, b: str, sep: str):
    aslash = a.endswith(sep)
    bslash = b.startswith(sep)
    if aslash and bslash:
        return a[:-1] + b
    elif aslash or bslash:
        return a + b
    else:
        return a + sep + b


def merge_query(query1: str, query2: str) -> str:
    query1 = split_query(query1)
    query2 = split_query(query2)
    merged = list()
    names = list(query1) + list(query2)
    for name in names:
        value1 = query1.pop(name, None)
        value2 = query2.pop(name, None)
        if value1 and value2:
            merged.append((name, join_string(value1, value2, "/")))
        elif value1:
            merged.append((name, value1))
        elif value2:
            merged.append((name, value2))
    return join_query(merged)


def join_uri(
    root: Union[str, ParseResult], relative: Union[str, ParseResult], **parse_options
) -> ParseResult:
    if isinstance(root, str):
        root = parse_uri(root, **parse_options)
    if isinstance(relative, str):
        relative = parse_uri(relative, **parse_options)
    if root.params or relative.params:
        raise NotImplementedError()
    path = os.path.join(root.path, relative.path)
    query = merge_query(root.query, relative.query)
    return ParseResult(root.scheme, root.netloc, path, "", query, "")


def uri_as_string(uri: Union[str, ParseResult]) -> Optional[str]:
    if isinstance(uri, str):
        return uri
    elif isinstance(uri, ParseResult):
        return uri.geturl()



from time import sleep as _sleep  # hide from task discovery
from ewokscore.taskwithprogress import TaskWithProgress


class SumList(
    TaskWithProgress,
    input_names=["list"],
    optional_input_names=["delay"],
    output_names=["sum"],
):
    """Add items from a list"""

    def run(self):
        if self.inputs.list is None:
            raise ValueError("list should be provided")
        if self.inputs.delay:
            delay = self.inputs.delay
        else:
            delay = 0
        sum_ = 0
        n_elmt = len(self.inputs.list)
        for i_elmt, elmt in enumerate(self.inputs.list):
            sum_ += elmt
            self.progress = (i_elmt / n_elmt) * 100.0
            _sleep(delay)
        self.progress = 100.0
        self.outputs.sum = sum_



from functools import partial
from typing import Optional, Tuple
import warnings

from ewoksutils.import_utils import import_method
from ewoksutils.import_utils import import_qualname

from .task import Task
from .methodtask import MethodExecutorTask
from .scripttask import ScriptExecutorTask
from .ppftasks import PpfMethodExecutorTask
from .ppftasks import PpfPortTask
from .notebooktask import NotebookExecutorTask
from .dynamictask import get_dynamically_task_class
from .node import get_node_label
from .node import NodeIdType


TASK_EXECUTABLE_ATTRIBUTE = (
    "class",
    "method",
    "ppfmethod",
    "ppfport",
    "script",
    "notebook",
    "task",
    "task_type",
)

TASK_EXECUTABLE_ATTRIBUTE_ALL = TASK_EXECUTABLE_ATTRIBUTE + ("graph",)

TASK_EXECUTABLE_ATTRIBUTE_STR = (
    ", ".join(map(repr, TASK_EXECUTABLE_ATTRIBUTE[:-1]))
    + " or "
    + repr(TASK_EXECUTABLE_ATTRIBUTE[-1])
)

TASK_EXECUTABLE_ATTRIBUTE_ALL_STR = (
    ", ".join(map(repr, TASK_EXECUTABLE_ATTRIBUTE_ALL[:-1]))
    + " or "
    + repr(TASK_EXECUTABLE_ATTRIBUTE_ALL[-1])
)
TASK_EXECUTABLE_ERROR_MSG = f"Task{{}} requires the {TASK_EXECUTABLE_ATTRIBUTE_STR} key"
TASK_EXECUTABLE_ERROR_MSG_ALL = (
    f"Task{{}} requires the {TASK_EXECUTABLE_ATTRIBUTE_ALL_STR} key"
)


def raise_task_error(node_label: str, all: bool = True):
    if node_label:
        node_label = " " + repr(node_label)
    if all:
        error_fmt = TASK_EXECUTABLE_ERROR_MSG_ALL
    else:
        error_fmt = TASK_EXECUTABLE_ERROR_MSG
    raise ValueError(error_fmt.format(node_label))


def task_executable_info(
    node_id: NodeIdType, node_attrs: dict, all: bool = False
) -> Tuple[str, dict]:
    if all:
        keys = TASK_EXECUTABLE_ATTRIBUTE_ALL
    else:
        keys = TASK_EXECUTABLE_ATTRIBUTE
    key = node_attrs.keys() & set(keys)
    if len(key) != 1:
        node_label = get_node_label(node_id, node_attrs)
        raise_task_error(node_label, all=all)
    key = key.pop()

    if key == "task_type":
        task_type = node_attrs[key]
    else:
        warnings.warn(
            f"'{key}' is deprecated in favor of 'task_type' with 'task_identifier'",
            DeprecationWarning,
        )
        value = node_attrs.pop(key)
        if key != "ppfport":
            node_attrs["task_identifier"] = value
        if key == "task":
            task_type = "generated"
        else:
            task_type = key
        node_attrs["task_type"] = task_type

    has_task_generator = bool(node_attrs.get("task_generator"))
    if task_type == "generated":
        if not has_task_generator:
            raise ValueError("node attribute 'task_generator' is missing")
    elif has_task_generator:
        raise ValueError(
            "node attribute 'task_generator' should only be specified when 'task_type' is 'generated'"
        )

    has_task_identifier = bool(node_attrs.get("task_identifier"))
    if task_type == "ppfport":
        if has_task_identifier:
            raise ValueError(
                "node attribute 'task_identifier' should not be used when 'task_type' is 'ppfport'"
            )
    elif not has_task_identifier:
        raise ValueError("node attribute 'task_identifier' is missing")

    info = dict()
    if task_type != "ppfport":
        info["task_identifier"] = node_attrs["task_identifier"]
    if task_type == "generated":
        info["task_generator"] = node_attrs["task_generator"]
    return task_type, info


def validate_task_executable(node_id: NodeIdType, node_attrs: dict, all: bool = False):
    task_executable_info(node_id, node_attrs, all=all)


def instantiate_task(
    node_id: NodeIdType,
    node_attrs: dict,
    inputs: Optional[dict] = None,
    varinfo: Optional[dict] = None,
    execinfo: Optional[dict] = None,
) -> Task:
    """
    :param node_id:
    :param node_attrs: node attributes of the graph representation
    :param inputs: dynamic inputs (from other tasks)
    :param varinfo: `Variable` constructor arguments
    :param execinfo:
    :returns Task:
    """
    # Default inputs
    task_inputs = node_attrs.get("default_inputs", list())
    task_inputs = {d["name"]: d["value"] for d in task_inputs}
    # Dynamic inputs (from other tasks)
    if inputs:
        task_inputs.update(inputs)

    # Instantiate task
    task_type, task_info = task_executable_info(node_id, node_attrs)
    task_kwargs = {
        "inputs": task_inputs,
        "varinfo": varinfo,
        "node_id": node_id,
        "node_attrs": node_attrs,
        "execinfo": execinfo,
    }
    if task_type == "class":
        return Task.instantiate(task_info["task_identifier"], **task_kwargs)

    if task_type == "method":
        task_inputs[MethodExecutorTask.METHOD_ARGUMENT] = task_info["task_identifier"]
        return MethodExecutorTask(**task_kwargs)

    if task_type == "ppfmethod":
        task_inputs[PpfMethodExecutorTask.METHOD_ARGUMENT] = task_info[
            "task_identifier"
        ]
        return PpfMethodExecutorTask(**task_kwargs)

    if task_type == "ppfport":
        return PpfPortTask(**task_kwargs)

    if task_type == "script":
        task_inputs[ScriptExecutorTask.SCRIPT_ARGUMENT] = task_info["task_identifier"]
        return ScriptExecutorTask(**task_kwargs)

    if task_type == "method":
        task_inputs[NotebookExecutorTask.NOTEBOOK_ARGUMENT] = task_info[
            "task_identifier"
        ]
        return NotebookExecutorTask(**task_kwargs)

    if task_type == "generated":
        task_class = get_dynamically_task_class(
            node_attrs.get("task_generator"), task_info["task_identifier"]
        )
        return task_class(**task_kwargs)

    node_label = get_node_label(node_id, node_attrs)
    raise_task_error(node_label, all=False)


def add_dynamic_inputs(
    dynamic_inputs: dict,
    link_attrs: dict,
    source_results: dict,
    source_id: Optional[NodeIdType] = None,
    target_id: Optional[NodeIdType] = None,
):
    err_suffix = ""
    if source_id or target_id:
        err_suffix = f" (Link '{source_id}' -> '{target_id}')"
    map_all_data = link_attrs.get("map_all_data", False)
    data_mapping = link_attrs.get("data_mapping", list())
    if map_all_data and data_mapping:
        raise ValueError(
            f"'data_mapping' and 'map_all_data' cannot be used together{err_suffix}"
        )
    if map_all_data:
        data_mapping = [{"target_input": s, "source_output": s} for s in source_results]
        for from_arg in source_results:
            to_arg = from_arg
            dynamic_inputs[to_arg] = source_results[from_arg]
    for arg in data_mapping:
        output_arg = arg.get("source_output")
        try:
            input_arg = arg["target_input"]
        except KeyError:
            raise KeyError(
                f"Argument '{arg}' is missing an 'input' key{err_suffix}"
            ) from None
        if output_arg:
            try:
                dynamic_inputs[input_arg] = source_results[output_arg]
            except KeyError:
                raise KeyError(
                    f"'{output_arg}' is not an output variable of the source node{err_suffix}"
                ) from None
        else:
            dynamic_inputs[input_arg] = source_results


def task_executable(node_id: NodeIdType, node_attrs: dict):
    task_type, task_info = task_executable_info(node_id, node_attrs)
    if task_type == "class":
        return task_info["task_identifier"], import_qualname
    if task_type == "method":
        return task_info["task_identifier"], import_method
    if task_type == "ppfmethod":
        return task_info["task_identifier"], import_method
    if task_type == "ppfport":
        return None, None
    if task_type == "script":
        return task_info["task_identifier"], None
    if task_type == "notebook":
        return task_info["task_identifier"], None
    if task_type == "generated":
        return task_info["task_identifier"], partial(
            get_dynamically_task_class, node_attrs.get("task_generator")
        )
    node_label = get_node_label(node_id, node_attrs)
    raise_task_error(node_label, all=False)


def get_task_class(node_id: NodeIdType, node_attrs: dict):
    task_type, task_info = task_executable_info(node_id, node_attrs)
    if task_type == "class":
        return Task.get_subclass(task_info["task_identifier"])
    if task_type == "method":
        return MethodExecutorTask
    if task_type == "ppfmethod":
        return PpfMethodExecutorTask
    if task_type == "ppfport":
        return PpfPortTask
    if task_type == "script":
        return ScriptExecutorTask
    if task_type == "notebook":
        return NotebookExecutorTask
    if task_type == "task":
        return get_dynamically_task_class(
            node_attrs.get("task_generator"), task_info["task_identifier"]
        )

    node_label = get_node_label(node_id, node_attrs)
    raise_task_error(node_label, all=False)



from . import graph


@graph
def acyclic2():
    task = "ewokscore.tests.examples.tasks.errorsumtask.ErrorSumTask"
    nodes = [
        {
            "id": "task1",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task2",
            "default_inputs": [
                {"name": "b", "value": 2},
                {"name": "raise_error", "value": True},
            ],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task3",
            "default_inputs": [{"name": "b", "value": 3}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task4",
            "default_inputs": [{"name": "a", "value": 3}, {"name": "b", "value": 4}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task5",
            "default_inputs": [{"name": "b", "value": 5}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task6",
            "default_inputs": [{"name": "b", "value": 6}],
            "task_type": "class",
            "task_identifier": task,
        },
    ]

    links = [
        {
            "source": "task1",
            "target": "task2",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task3",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task4",
            "on_error": True,
        },
        {
            "source": "task3",
            "target": "task5",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task4",
            "target": "task6",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
    ]

    graph = {
        "links": links,
        "nodes": nodes,
    }

    expected_results = {
        "task1": {"result": 1},
        "task2": None,  # error
        "task3": None,  # error branch
        "task4": {"result": 7},
        "task5": None,  # error branch
        "task6": {"result": 13},
    }

    return graph, expected_results



import random
import hashlib
from typing import Any, Optional, Type, Union
from collections.abc import Mapping, Iterable, Set
import numpy
from ewoksutils.import_utils import qualname
from . import missing_data


def classhashdata(cls: Type) -> bytes:
    return qualname(cls).encode()


def multitype_sorted(sequence: Iterable, key=None) -> list:
    try:
        return sorted(sequence, key=key)
    except TypeError:
        pass
    if key is None:

        def key(item):
            return item

    adict = dict()
    for item in sequence:
        typename = type(key(item)).__name__
        adict.setdefault(typename, list()).append(item)

    return [
        item
        for _, items in sorted(adict.items(), key=lambda tpl: tpl[0])
        for item in sorted(items, key=key)
    ]


class UniversalHash:
    def __init__(self, hexdigest: Union[str, bytes]):
        if isinstance(hexdigest, bytes):
            hexdigest = hexdigest.decode()
        if not isinstance(hexdigest, str):
            raise TypeError(hexdigest, type(hexdigest))
        self._hexdigest = hexdigest

    def __hash__(self):
        # make it python hashable (to use in sets and dict keys)
        return hash(self._hexdigest)

    def __repr__(self):
        return "UniversalHash('{}')".format(self)

    def __str__(self):
        return self._hexdigest

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)


def uhash(value, _hash=None) -> UniversalHash:
    """Universial hash (as opposed to python's `hash`)."""
    # Avoid using python's hash!
    bdigest = _hash is None
    if bdigest:
        _hash = hashlib.sha256()
    _hash.update(classhashdata(type(value)))
    if value is None:
        pass
    elif isinstance(value, HasUhash):
        _hash.update(repr(value.uhash).encode())
    elif isinstance(value, UniversalHash):
        _hash.update(repr(value).encode())
    elif isinstance(value, bytes):
        _hash.update(value)
    elif isinstance(value, str):
        _hash.update(value.encode())
    elif isinstance(value, int):
        _hash.update(hex(value).encode())
    elif isinstance(value, float):
        _hash.update(value.hex().encode())
    elif isinstance(value, (numpy.ndarray, numpy.number)):
        _hash.update(value.tobytes())
    elif isinstance(value, Mapping):
        lst = multitype_sorted(value.items(), key=lambda item: item[0])
        if lst:
            keys, values = zip(*lst)
        else:
            keys = values = list()
        uhash(keys, _hash=_hash)
        uhash(values, _hash=_hash)
    elif isinstance(value, Set):
        values = multitype_sorted(value)
        uhash(values, _hash=_hash)
    elif isinstance(value, Iterable):
        # Ordered
        for v in value:
            uhash(v, _hash=_hash)
    else:
        # TODO: register custom types
        raise TypeError(f"cannot uhash {value} (type: {type(value)})")
    if bdigest:
        return UniversalHash(_hash.hexdigest())


class HasUhash:
    @property
    def uhash(self) -> Optional[UniversalHash]:
        raise NotImplementedError

    def __hash__(self):
        # make it python hashable (to use in sets and dict keys)
        uhash = self.uhash
        if uhash is None:
            return hash(id(self))
        else:
            return hash(uhash)

    def __eq__(self, other):
        if isinstance(other, HasUhash):
            uhash = other.uhash
        elif isinstance(other, UniversalHash):
            uhash = other
        else:
            raise TypeError(other, type(other))
        return self.uhash == uhash

    def _get_repr_data(self) -> dict:
        data = dict()
        uhash = self.uhash
        if uhash is None:
            data["uhash"] = None
        else:
            data["uhash"] = repr(str(uhash))
        return data

    def __repr__(self):
        data = self._get_repr_data()
        if data:
            sdata = ", ".join([f"{k}={v}" for k, v in data.items()])
            return f"{super().__repr__()}({sdata})"
        else:
            return super().__repr__()

    def __str__(self):
        data = self._get_repr_data()
        if data:
            sdata = ", ".join([f"{k}={v}" for k, v in data.items()])
            return f"{qualname(type(self))}({sdata})"
        else:
            return qualname(type(self))


PreUhashTypes = Union[str, bytes, UniversalHash, HasUhash]


class UniversalHashable(HasUhash):
    """The universal hash of an instance of this class is based on:

     * pre-uhash
     * instance nonce (if any)

    The universal hash is equal to the pre-hash when an instance nonce is not provided.

    The pre-uhash is either provided or based on:

     * data
     * class nonce (class qualifier name, class version, superclass nonce)
    """

    __CLASS_NONCE = None
    __VERSION = None
    MISSING_DATA = missing_data.MISSING_DATA

    def __init__(
        self,
        pre_uhash: Optional[PreUhashTypes] = None,
        instance_nonce: Optional[Any] = None,
    ):
        self.set_uhash_init(pre_uhash=pre_uhash, instance_nonce=instance_nonce)

    def __init_subclass__(subcls, version=None, **kwargs):
        super().__init_subclass__(**kwargs)
        supercls_data = subcls.class_nonce()
        subcls.__VERSION = version
        subcls_data = subcls.class_nonce_data()
        subcls.__CLASS_NONCE = str(uhash((subcls_data, supercls_data)))

    def set_uhash_init(
        self,
        pre_uhash: Optional[PreUhashTypes] = None,
        instance_nonce: Optional[Any] = None,
    ):
        self.__set_pre_uhash(pre_uhash)
        self.__original_pre_uhash = self.__pre_uhash
        self.__instance_nonce = instance_nonce
        self.__original__instance_nonce = instance_nonce

    def get_uhash_init(self, serialize=False):
        pre_uhash = self.__original_pre_uhash
        if serialize:
            if isinstance(pre_uhash, HasUhash):
                pre_uhash = str(pre_uhash.uhash)
            elif isinstance(pre_uhash, UniversalHash):
                pre_uhash = str(pre_uhash)
        return {
            "pre_uhash": pre_uhash,
            "instance_nonce": self.__original__instance_nonce,
        }

    def __set_pre_uhash(self, pre_uhash):
        if pre_uhash is None:
            self.__pre_uhash = None
        elif isinstance(pre_uhash, (str, bytes)):
            self.__pre_uhash = UniversalHash(pre_uhash)
        elif isinstance(pre_uhash, (UniversalHash, HasUhash)):
            self.__pre_uhash = pre_uhash
        else:
            self.__pre_uhash = uhash(pre_uhash)

    @classmethod
    def class_nonce(cls):
        return cls.__CLASS_NONCE

    @classmethod
    def class_nonce_data(cls):
        return qualname(cls), cls.__VERSION

    def instance_nonce(self):
        return self.__instance_nonce

    def fix_uhash(self):
        """Fix the uhash when it is derived from the uhash data."""
        if self.__pre_uhash is not None:
            return
        keep, self.__instance_nonce = self.__instance_nonce, None
        try:
            pre_uhash = self.uhash
        finally:
            self.__instance_nonce = keep
        self.__set_pre_uhash(pre_uhash)

    def undo_fix_uhash(self):
        self.__pre_uhash = self.__original_pre_uhash

    def cleanup_references(self):
        """Remove all references to other hashables.
        Side effect: fixes the uhash when it depends on another hashable.
        """
        if isinstance(self.__pre_uhash, HasUhash):
            pre_uhash = self.__pre_uhash.uhash
            self.__pre_uhash = pre_uhash
            self.__original_pre_uhash = pre_uhash

    @property
    def uhash(self) -> Optional[UniversalHash]:
        _uhash = self.__pre_uhash
        if _uhash is None:
            data = self._uhash_data()
            if missing_data.is_missing_data(data):
                return None
            cnonce = self.class_nonce()
            inonce = self.instance_nonce()
            if inonce is None:
                return uhash((data, cnonce))
            else:
                return uhash((data, cnonce, inonce))
        else:
            if isinstance(_uhash, HasUhash):
                _uhash = _uhash.uhash
                if _uhash is None:
                    return None
            inonce = self.instance_nonce()
            if inonce is None:
                return _uhash
            else:
                return uhash((_uhash, inonce))

    def _uhash_data(self):
        return self.MISSING_DATA

    def uhash_randomize(self):
        self.__instance_nonce = random.randint(-1e100, 1e100)

    def undo_randomize(self):
        self.__instance_nonce = self.__original__instance_nonce



from time import sleep as _sleep  # hide from task discovery
from ewokscore.taskwithprogress import TaskWithProgress


class SumTask(
    TaskWithProgress,
    input_names=["a"],
    optional_input_names=["b", "delay"],
    output_names=["result"],
):
    """Add two numbers with a delay"""

    def run(self):
        result = self.inputs.a
        if self.inputs.b:
            result += self.inputs.b
        self.progress = 0
        if self.inputs.delay:
            dt = self.inputs.delay / 100
            for i in range(100):
                _sleep(dt)
                self.progress = i + 1
        else:
            self.progress = 100
        self.outputs.result = result



from ewoksutils.logging_utils.sqlite3 import Sqlite3Handler
from .base import EwoksEventHandlerMixIn


class Sqlite3EwoksEventHandler(EwoksEventHandlerMixIn, Sqlite3Handler):
    def __init__(self, uri: str):
        super().__init__(uri=uri, table="ewoks_events", field_types=self.FIELD_TYPES)



from ewoksutils.import_utils import qualname
from ewokscore import Task
from ewokscore.inittask import instantiate_task


def task_class_generator(qualname):
    registry_name = qualname
    if registry_name in Task.get_subclass_names():
        return Task.get_subclass(qualname)

    class DynamicTask(Task, output_names=["result"], registry_name=registry_name):
        def run(self):
            self.outputs.result = qualname

    return DynamicTask


def test_task_class_generator():
    task_name = "some.unique.task.name"
    task = instantiate_task(
        "node_id",
        {
            "task_type": "generated",
            "task_identifier": task_name,
            "task_generator": qualname(task_class_generator),
        },
    )
    task.execute()
    assert task.get_output_values() == {"result": task_name}



from typing import Hashable, Optional, Union
import networkx
from ewoksutils.import_utils import qualname

from .. import inittask
from .subgraph import extract_graph_nodes
from .subgraph import add_subgraph_links
from . import analysis
from . import serialize
from .validate import validate_graph
from .multigraph import flatten_multigraph
from .error_handlers import connect_default_error_handlers
from .execute.sequential import execute_graph
from .compare import graphs_are_equal


def load_graph(source=None, representation=None, **load_options):
    if isinstance(source, TaskGraph):
        return source
    else:
        return TaskGraph(source=source, representation=representation, **load_options)


def node_has_links(graph, node_id):
    try:
        next(graph.successors(node_id))
    except StopIteration:
        try:
            next(graph.predecessors(node_id))
        except StopIteration:
            return False
    return True


def merge_graphs(graphs, graph_attrs=None, rename_nodes=None, **load_options):
    lst = list()
    if rename_nodes is None:
        rename_nodes = [True] * len(graphs)
    else:
        assert len(graphs) == len(rename_nodes)
    for g, rename in zip(graphs, rename_nodes):
        g = load_graph(g, **load_options)
        gname = repr(g)
        g = g.graph
        if rename:
            mapping = {s: (gname, s) for s in g.nodes}
            g = networkx.relabel_nodes(g, mapping, copy=True)
        lst.append(g)
    ret = load_graph(networkx.compose_all(lst), **load_options)
    if graph_attrs:
        ret.graph.graph.update(graph_attrs)
    return ret


def get_subgraphs(graph: networkx.DiGraph, **load_options):
    subgraphs = dict()
    for node_id, node_attrs in graph.nodes.items():
        task_type, task_info = inittask.task_executable_info(
            node_id, node_attrs, all=True
        )
        if task_type == "graph":
            g = load_graph(task_info["task_identifier"], **load_options)
            g.graph.graph["id"] = node_id
            node_label = node_attrs.get("label")
            if node_label:
                g.graph.graph["label"] = node_label
            subgraphs[node_id] = g
    return subgraphs


class TaskGraph:
    """The API for graph analysis is provided by `networkx`.
    Any directed graph is supported (cyclic or acyclic).

    Loop over the dependencies of a task

    .. code-block:: python

        for source in taskgraph.predecessors(target):
            link_attrs = taskgraph.graph[source][target]

    Loop over the tasks dependent on a task

    .. code-block:: python

        for target in taskgraph.successors(source):
            link_attrs = taskgraph.graph[source][target]

    For acyclic graphs, sequential task execution can be done like this:

    .. code-block:: python

        taskgraph.execute()
    """

    def __init__(self, source=None, representation=None, **load_options):
        self.load(source=source, representation=representation, **load_options)

    def __repr__(self):
        return self.graph_label

    @property
    def graph_id(self) -> Hashable:
        return self.graph.graph.get("id", qualname(type(self)))

    @property
    def graph_label(self) -> str:
        return str(self.graph.graph.get("label", self.graph_id))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(other, type(other))
        return graphs_are_equal(self.graph, other.graph)

    def load(
        self,
        source=None,
        representation: Optional[Union[serialize.GraphRepresentation, str]] = None,
        subgraph_representation: Optional[
            Union[serialize.GraphRepresentation, str]
        ] = None,
        **load_options,
    ) -> None:
        graph = serialize.load(
            source=source, representation=representation, **load_options
        )

        if subgraph_representation is not None:
            representation = subgraph_representation

        subgraphs = get_subgraphs(graph, representation=representation, **load_options)
        if subgraphs:
            # Extract
            edges, update_attrs = extract_graph_nodes(graph, subgraphs)
            graph = flatten_multigraph(graph)

            # Merged
            self.graph = graph
            graphs = [self] + list(subgraphs.values())
            rename_nodes = [False] + [True] * len(subgraphs)
            graph = merge_graphs(
                graphs,
                graph_attrs=graph.graph,
                rename_nodes=rename_nodes,
                **load_options,
            ).graph

            # Re-link
            add_subgraph_links(graph, edges, update_attrs)

        graph = flatten_multigraph(graph)
        connect_default_error_handlers(graph)
        validate_graph(graph)
        self.graph = graph

    def dump(
        self,
        destination=None,
        representation: Optional[Union[serialize.GraphRepresentation, str]] = None,
        **kw,
    ) -> Optional[Union[str, dict]]:
        return serialize.dump(
            self.graph, destination=destination, representation=representation, **kw
        )

    def serialize(self) -> str:
        return self.dump(representation=serialize.GraphRepresentation.json_string)

    @property
    def is_cyclic(self) -> bool:
        return analysis.graph_is_cyclic(self.graph)

    @property
    def has_conditional_links(self) -> bool:
        return analysis.graph_has_conditional_links(self.graph)

    def execute(self, *args, **kw):
        return execute_graph(self.graph, *args, **kw)



import pytest
from silx.io.dictdump import nxtodict
from ewokscore.hashing import UniversalHashable, uhash
from ewokscore.persistence.json import JsonProxy
from ewokscore.persistence.nexus import NexusProxy
from ewokscore.persistence import instantiate_data_proxy


def test_json_proxy_uri(tmpdir):
    hashable = UniversalHashable(uhash("somedata"))
    identifier = str(hashable.uhash)
    proxy = JsonProxy()
    assert proxy.uri is None
    proxy = JsonProxy(uhash_source=hashable)
    assert proxy.uri is None

    proxy = JsonProxy(uhash_source=hashable, root_uri=str(tmpdir))
    assert str(proxy.uri) == f"json://{tmpdir/identifier}.json"
    tmpfile = tmpdir / "file"
    proxy = JsonProxy(uhash_source=hashable, root_uri=f"{tmpfile}.json")
    assert str(proxy.uri) == f"json://{tmpfile/identifier}.json"
    proxy = JsonProxy(uhash_source=hashable, root_uri=f"{tmpfile}.json?path=/a")
    assert str(proxy.uri) == f"json://{tmpfile/'a'/identifier}.json"
    proxy = JsonProxy(uhash_source=hashable, root_uri=f"{tmpfile}.json?path=/a/b")
    assert str(proxy.uri) == f"json://{tmpfile/'a'/'b'/identifier}.json"
    proxy = JsonProxy(uhash_source=hashable, root_uri=f"{tmpfile}.json?path=/a/b/c")
    assert str(proxy.uri) == f"json://{tmpfile/'a'/'b'/'c'/identifier}.json"

    proxy2 = JsonProxy(proxy.uri)
    assert proxy.uri == proxy2.uri
    assert str(proxy2.uri) == f"json://{tmpfile/'a'/'b'/'c'/identifier}.json"


def test_nexus_proxy_uri(tmpdir):
    hashable = UniversalHashable(uhash("somedata"))
    identifier = str(hashable.uhash)
    proxy = NexusProxy()
    assert proxy.uri is None
    proxy = NexusProxy(uhash_source=hashable)
    assert proxy.uri is None

    proxy = NexusProxy(uhash_source=hashable, root_uri=str(tmpdir))
    assert str(proxy.uri) == f"nexus://{tmpdir/identifier}.nx?path={identifier}"
    tmpfile = tmpdir / "file"
    proxy = NexusProxy(uhash_source=hashable, root_uri=f"{tmpfile}.nx")
    assert str(proxy.uri) == f"nexus://{tmpfile}.nx?path={identifier}"
    proxy = NexusProxy(uhash_source=hashable, root_uri=f"{tmpfile}.h5?path=/a")
    assert str(proxy.uri) == f"nexus://{tmpfile}.h5?path=a/{identifier}"
    proxy = NexusProxy(uhash_source=hashable, root_uri=f"{tmpfile}.nx?path=/a/b")
    assert str(proxy.uri) == f"nexus://{tmpfile}.nx?path=a/b/{identifier}"
    proxy = NexusProxy(uhash_source=hashable, root_uri=f"{tmpfile}.nx?path=/a/b/c")
    assert str(proxy.uri) == f"nexus://{tmpfile}.nx?path=a/b/c/{identifier}"

    proxy2 = JsonProxy(proxy.uri)
    assert proxy.uri == proxy2.uri
    assert str(proxy2.uri) == f"nexus://{tmpfile}.nx?path=a/b/c/{identifier}"


@pytest.mark.parametrize("scheme", ("json", "nexus"))
@pytest.mark.parametrize("full", (True, False))
def test_proxy_dump(scheme, full, tmpdir):
    if scheme == "nexus":
        extension = ".nx"
    else:
        extension = ".json"

    root_uri = tmpdir
    if full:
        root_uri /= f"dataset{extension}::/scan/task/output_variable_a"
    root_uri = f"{scheme}://{root_uri}"

    hashable = UniversalHashable(uhash("somedata"))
    proxy = instantiate_data_proxy(
        scheme=scheme, uhash_source=hashable, root_uri=root_uri
    )
    data = [1, 2, 3]
    proxy.dump(data)

    proxy2 = instantiate_data_proxy(
        scheme=scheme, uhash_source=hashable, root_uri=root_uri
    )
    data2 = proxy2.load()
    if scheme == "nexus":
        data2 = data2.tolist()
        adict = nxtodict(str(proxy2.path))
        assert adict["@NX_class"] == "NXroot"
        if full:
            adict = adict["scan"]
            assert adict["@NX_class"] == "NXentry"
            adict = adict["task"]
            assert adict["@NX_class"] == "NXprocess"
            adict = adict["output_variable_a"]
            assert adict["@NX_class"] == "NXcollection"
    assert data == data2



from ewokscore.graph import load_graph
from ewokscore.node import get_node_label


def subsubmodel1():
    nodes = [
        {
            "id": "a",
            "label": "task3",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "b",
            "label": "task4",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "c",
            "label": "special_task5",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "special_handler",
            "label": "special_handler3",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "graph_handler",
            "label": "graph_handler3",
            "task_type": "method",
            "task_identifier": "dummy",
            "default_error_node": True,
        },
    ]
    links = [
        {"source": "a", "target": "b", "map_all_data": True},
        {"source": "b", "target": "c", "map_all_data": True},
        {
            "source": "c",
            "target": "special_handler",
            "map_all_data": True,
            "on_error": True,
        },
    ]
    return {"graph": {"id": "subsubmodel1"}, "nodes": nodes, "links": links}


def submodel1():
    nodes = [
        {
            "id": "a",
            "label": "task2",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {"id": "b", "task_type": "graph", "task_identifier": subsubmodel1()},
        {
            "id": "c",
            "label": "special_task6",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "special_handler",
            "label": "special_handler2",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "graph_handler",
            "label": "graph_handler2",
            "task_type": "method",
            "task_identifier": "dummy",
            "default_error_node": True,
        },
    ]
    links = [
        {"source": "a", "target": "b", "sub_target": "a", "map_all_data": True},
        {"source": "b", "target": "c", "sub_source": "c", "map_all_data": True},
        {
            "source": "c",
            "target": "special_handler",
            "map_all_data": True,
            "on_error": True,
        },
    ]
    return {"graph": {"id": "submodel1"}, "nodes": nodes, "links": links}


def model1():
    nodes = [
        {
            "id": "a",
            "label": "task1",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {"id": "b", "task_type": "graph", "task_identifier": submodel1()},
        {
            "id": "c",
            "label": "special_task7",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "special_handler",
            "label": "special_handler1",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "graph_handler",
            "label": "graph_handler1",
            "task_type": "method",
            "task_identifier": "dummy",
            "default_error_node": True,
        },
    ]
    links = [
        {"source": "a", "target": "b", "sub_target": "a", "map_all_data": True},
        {"source": "b", "target": "c", "sub_source": "c", "map_all_data": True},
        {
            "source": "c",
            "target": "special_handler",
            "map_all_data": True,
            "on_error": True,
        },
    ]
    return {"graph": {"id": "model1"}, "nodes": nodes, "links": links}


def test_default_error_handlers1():
    graph = load_graph(model1()).graph

    links = dict()
    for (source_id, target_id), link_attrs in graph.edges.items():
        source_label = get_node_label(source_id, graph.nodes[source_id])
        target_label = get_node_label(target_id, graph.nodes[target_id])
        links[(source_label, target_label)] = link_attrs

    expected = {
        # normal connection
        ("task1", "task2"): {"map_all_data": True},
        ("task2", "task3"): {"map_all_data": True},
        ("task3", "task4"): {"map_all_data": True},
        ("task4", "special_task5"): {"map_all_data": True},
        ("special_task5", "special_task6"): {"map_all_data": True},
        ("special_task6", "special_task7"): {"map_all_data": True},
        # error handlers of special tasks
        ("special_task5", "special_handler3"): {"map_all_data": True, "on_error": True},
        ("special_task6", "special_handler2"): {"map_all_data": True, "on_error": True},
        ("special_task7", "special_handler1"): {"map_all_data": True, "on_error": True},
        # error handlers of normal tasks
        ("task1", "graph_handler1"): {"map_all_data": True, "on_error": True},
        ("task2", "graph_handler2"): {"map_all_data": True, "on_error": True},
        ("task3", "graph_handler3"): {"map_all_data": True, "on_error": True},
        ("task4", "graph_handler3"): {"map_all_data": True, "on_error": True},
        # error handlers of special handlers
        ("special_handler1", "graph_handler1"): {
            "map_all_data": True,
            "on_error": True,
        },
        ("special_handler2", "graph_handler2"): {
            "map_all_data": True,
            "on_error": True,
        },
        ("special_handler3", "graph_handler3"): {
            "map_all_data": True,
            "on_error": True,
        },
        # error handlers of graph handlers
        ("graph_handler2", "graph_handler1"): {"map_all_data": True, "on_error": True},
        ("graph_handler3", "graph_handler2"): {"map_all_data": True, "on_error": True},
    }

    assert expected == links


def submodel2():
    graph = {
        "id": "submodel2",
        "input_nodes": [{"id": "in", "node": "a"}],
        "output_nodes": [{"id": "out", "node": "c"}],
    }
    nodes = [
        {
            "id": "a",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "b",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "c",
            "task_type": "method",
            "task_identifier": "dummy",
        },
    ]
    links = [
        {"source": "a", "target": "b", "map_all_data": True},
        {"source": "b", "target": "c", "map_all_data": True},
    ]
    return {"graph": graph, "nodes": nodes, "links": links}


def model2():
    nodes = [
        {
            "id": "start",
            "task_type": "method",
            "task_identifier": "dummy",
        },
        {
            "id": "error_graph",
            "task_type": "graph",
            "task_identifier": submodel2(),
            "default_error_node": True,
        },
        {
            "id": "end",
            "task_type": "method",
            "task_identifier": "dummy",
        },
    ]
    links = [
        {"source": "start", "target": "end", "map_all_data": True},
    ]
    return {"graph": {"id": "model2"}, "nodes": nodes, "links": links}


def test_default_error_handlers2():
    graph = load_graph(model2()).graph

    links = dict()
    for (source_id, target_id), link_attrs in graph.edges.items():
        source_label = get_node_label(source_id, graph.nodes[source_id])
        target_label = get_node_label(target_id, graph.nodes[target_id])
        links[(source_label, target_label)] = link_attrs

    expected = {
        ("start", "end"): {"map_all_data": True},
        ("start", "error_graph:a"): {"map_all_data": True, "on_error": True},
        ("end", "error_graph:a"): {"map_all_data": True, "on_error": True},
        ("error_graph:a", "error_graph:b"): {"map_all_data": True},
        ("error_graph:b", "error_graph:c"): {"map_all_data": True},
    }

    assert expected == links



import pkgutil
import importlib
from functools import wraps

ALL_GRAPHS = None


def _discover_graphs():
    if ALL_GRAPHS is None:
        for pkginfo in pkgutil.walk_packages(__path__, __name__ + "."):
            importlib.import_module(pkginfo.name)


def graph_names():
    _discover_graphs()
    return ALL_GRAPHS.keys()


def get_graph(name) -> tuple:
    _discover_graphs()
    return ALL_GRAPHS[name]


def graph(graph_method):
    global ALL_GRAPHS
    name = graph_method.__name__

    @wraps(graph_method)
    def wrapper():
        g, result = graph_method()
        attrs = g.setdefault("graph", dict())
        attrs.setdefault("id", name)
        attrs.setdefault("label", name)
        attrs.setdefault("schema_version", "1.0")
        return g, result

    if ALL_GRAPHS is None:
        ALL_GRAPHS = dict()
    ALL_GRAPHS[name] = wrapper()
    return wrapper



import itertools
from typing import Iterable, Optional, Tuple, Union, Any
import networkx
from copy import deepcopy

from ..utils import dict_merge
from ..node import flatten_node_id


NodeIdType = Union[str, Tuple[str, Any]]  # Any is NodeIdType


def _append_subnode_id(node_id: NodeIdType, sub_node_id: str) -> NodeIdType:
    if isinstance(node_id, tuple):
        parent, child = node_id
        return parent, _append_subnode_id(child, sub_node_id)
    else:
        return node_id, sub_node_id


def _get_subgraph(node_id: NodeIdType, subgraphs: dict):
    if isinstance(node_id, str):
        return subgraphs.get(node_id)
    subgraph_id, subnode_id = node_id
    try:
        subgraph = subgraphs[subgraph_id]
    except KeyError:
        raise ValueError(node_id, f"{repr(subgraph_id)} is not a subgraph")
    flat_subnode_id = flatten_node_id(subnode_id)
    n = len(flat_subnode_id)
    for node_id in subgraph.graph.nodes:
        flat_node_id = flatten_node_id(node_id)
        nflatid = len(flat_node_id)
        if flat_node_id == flat_subnode_id:
            return None  # a task node
        if nflatid > n and flat_node_id[:n] == flat_subnode_id:
            return subgraph  # a graph node
    raise ValueError(
        f"{subnode_id} is not a node or subgraph of subgraph {repr(subgraph_id)}",
    )


def _alias_to_node_id(alias_attrs: dict) -> NodeIdType:
    sub_node = alias_attrs.get("sub_node", None)
    if sub_node:
        return alias_attrs["node"], sub_node
    else:
        return alias_attrs["node"]


def _resolve_node_aliases(
    node_id: NodeIdType, graph_attrs: dict, input_nodes: bool
) -> Iterable[Tuple[NodeIdType, dict]]:
    if input_nodes:
        aliases = graph_attrs.get("input_nodes", list())
    else:
        aliases = graph_attrs.get("output_nodes", list())
    aliases = [alias_attrs for alias_attrs in aliases if alias_attrs["id"] == node_id]
    if aliases:
        for alias_attrs in aliases:
            sub_node_id = _alias_to_node_id(alias_attrs)
            link_attributes = alias_attrs.get("link_attributes", dict())
            yield sub_node_id, link_attributes
    else:
        yield node_id, dict()


def _resolve_all_node_aliases(
    graph_attrs: dict, input_nodes: bool
) -> Iterable[NodeIdType]:
    if input_nodes:
        aliases = graph_attrs.get("input_nodes", list())
    else:
        aliases = graph_attrs.get("output_nodes", list())
    for alias_attrs in aliases:
        yield _alias_to_node_id(alias_attrs)


def _get_subnode_ids(
    node_id: NodeIdType, link_attrs_subgraph_keys: dict, subgraphs: dict, source: bool
) -> Iterable[Tuple[NodeIdType, Optional[dict]]]:
    if source:
        key = "sub_source"
    else:
        key = "sub_target"
    subgraph = _get_subgraph(node_id, subgraphs)
    if subgraph is None:
        # node_id is not a subgraph
        if key in link_attrs_subgraph_keys:
            raise ValueError(
                f"'{node_id}' is not a graph so 'sub_source' should not be specified"
            )
        yield node_id, None
    else:
        # node_id is a subgraph
        try:
            sub_node_id = link_attrs_subgraph_keys[key]
        except KeyError:
            raise ValueError(
                f"the '{key}' attribute to specify a node in subgraph '{node_id}' is missing"
            ) from None
        for sub_node_id, link_attributes in _resolve_node_aliases(
            sub_node_id, subgraph.graph.graph, input_nodes=not source
        ):
            new_node_id = _append_subnode_id(node_id, sub_node_id)
            yield new_node_id, link_attributes


def _get_subnode_attributes(
    node_id: NodeIdType, subgraphs: dict, graph_node_attrs: dict
) -> Iterable[Tuple[NodeIdType, dict]]:
    """Update all input node attributes of the subgraph with the graph node attributes from the super graph"""
    transfer_attributes = {
        "default_inputs",
        "inputs_complete",
        "conditions_else_value",
        "default_error_node",
    }
    node_attrs = {k: v for k, v in graph_node_attrs.items() if k in transfer_attributes}
    if not node_attrs:
        return
    subgraph = _get_subgraph(node_id, subgraphs)
    if subgraph is None:
        # node_id is not a subgraph
        return
    # node_id is a subgraph
    for sub_node_id in _resolve_all_node_aliases(
        subgraph.graph.graph, input_nodes=True
    ):
        new_node_id = _append_subnode_id(node_id, sub_node_id)
        yield new_node_id, node_attrs


def _get_subnode_links(
    source_id: NodeIdType,
    target_id: NodeIdType,
    link_attrs_subgraph_keys: dict,
    subgraphs: dict,
) -> Iterable[Tuple[NodeIdType, NodeIdType, dict, bool]]:
    sources = list(
        _get_subnode_ids(source_id, link_attrs_subgraph_keys, subgraphs, source=True)
    )
    targets = list(
        _get_subnode_ids(target_id, link_attrs_subgraph_keys, subgraphs, source=False)
    )
    for sub_source, source_link_attrs in sources:
        for sub_target, target_link_attrs in targets:
            if source_link_attrs:
                link_attrs = source_link_attrs
            else:
                link_attrs = dict()
            if target_link_attrs:
                link_attrs.update(target_link_attrs)
            target_is_graph = target_link_attrs is not None
            yield sub_source, sub_target, link_attrs, target_is_graph


def _replace_aliases(
    graph: networkx.DiGraph, subgraphs: dict, input_nodes: bool
) -> dict:
    if input_nodes:
        aliases = graph.graph.get("input_nodes")
        if not aliases:
            return
        source = False
        key = "sub_target"
    else:
        aliases = graph.graph.get("output_nodes")
        if not aliases:
            return
        source = True
        key = "sub_source"

    new_aliases = list()
    for alias_attrs in aliases:
        node_id = alias_attrs.get("node")
        if node_id is None:
            continue
        sub_node = alias_attrs.pop("sub_node", None)
        if sub_node:
            node_id = node_id, sub_node
        if not isinstance(node_id, tuple):
            new_aliases.append(alias_attrs)
            continue
        parent, child = node_id
        original_alias_attrs = alias_attrs
        for node_id, link_attrs in _get_subnode_ids(
            parent, {key: child}, subgraphs=subgraphs, source=source
        ):
            alias_attrs = deepcopy(original_alias_attrs)
            if link_attrs:
                link_attrs.update(alias_attrs.get("link_attributes", dict()))
                alias_attrs["link_attributes"] = link_attrs
            alias_attrs["node"] = node_id
            new_aliases.append(alias_attrs)

    if input_nodes:
        graph.graph["input_nodes"] = new_aliases
    else:
        graph.graph["output_nodes"] = new_aliases


def extract_graph_nodes(graph: networkx.DiGraph, subgraphs) -> Tuple[list, dict]:
    """Removes all graph nodes from `graph` and returns a list of edges
    between the nodes from `graph` and `subgraphs`.

    Nodes in sub-graphs are defines in the `link_attrs_subgraph_keys` link attribute.
    For example:

    .. code: python

        link_attrs = {
            "source": "subgraph1",
            "target": "subgraph2",
            "data_mapping": [{"source_output":"return_value", "target_input": 0}],
            "link_attrs_subgraph_keys": {
                "sub_source": ("subsubgraph", ("subsubsubgraph", "task2")),
                "sub_target": "task1",
            },
        }
    """
    edges = list()
    update_attrs = dict()
    graph_is_multi = graph.is_multigraph()
    for subgraph_id in subgraphs:
        # Nodes to be updated after flattening the graph
        for node_id, graph_node_attrs in _get_subnode_attributes(
            subgraph_id, subgraphs, graph.nodes[subgraph_id]
        ):
            if node_id in update_attrs:
                update_attrs[node_id].update(graph_node_attrs)
            else:
                update_attrs[node_id] = graph_node_attrs
        # Links to be made after flattening the graph
        it1 = (
            (source_id, subgraph_id) for source_id in graph.predecessors(subgraph_id)
        )
        it2 = ((subgraph_id, target_id) for target_id in graph.successors(subgraph_id))
        for source_id, target_id in itertools.chain(it1, it2):
            all_link_attrs = graph[source_id][target_id]
            if graph_is_multi:
                all_link_attrs = all_link_attrs.values()
            else:
                all_link_attrs = [all_link_attrs]
            for link_attrs in all_link_attrs:
                link_attrs_subgraph_keys = {
                    key: link_attrs.pop(key)
                    for key in ["sub_source", "sub_target", "sub_target_attributes"]
                    if key in link_attrs
                }
                if not link_attrs_subgraph_keys:
                    continue
                original_link_attrs = link_attrs
                links = _get_subnode_links(
                    source_id, target_id, link_attrs_subgraph_keys, subgraphs
                )
                for source, target, default_link_attrs, target_is_graph in links:
                    link_attrs = deepcopy(original_link_attrs)
                    if default_link_attrs:
                        link_attrs = {**default_link_attrs, **link_attrs}
                    sub_target_attributes = link_attrs_subgraph_keys.get(
                        "sub_target_attributes", None
                    )
                    if sub_target_attributes:
                        if not target_is_graph:
                            raise ValueError(
                                f"'{target_id}' is not a graph so 'sub_target_attributes' should not be specified"
                            )
                        if target in update_attrs:
                            update_attrs[target].update(sub_target_attributes)
                        else:
                            update_attrs[target] = sub_target_attributes
                    edges.append((source, target, link_attrs))

    _replace_aliases(graph, subgraphs, input_nodes=True)
    _replace_aliases(graph, subgraphs, input_nodes=False)
    graph.remove_nodes_from(subgraphs.keys())
    return edges, update_attrs


def add_subgraph_links(graph: networkx.DiGraph, edges: list, update_attrs: dict):
    # Output from extract_graph_nodes
    for source, target, _ in edges:
        if source not in graph.nodes:
            raise ValueError(
                f"Source node {repr(source)} of link |{repr(source)} -> {repr(target)}| does not exist"
            )
        if target not in graph.nodes:
            raise ValueError(
                f"Target node {repr(target)} of link |{repr(source)} -> {repr(target)}| does not exist"
            )
    graph.add_edges_from(edges)  # This adds missing nodes
    for node, attrs in update_attrs.items():
        if attrs:
            node_attrs = graph.nodes[node]
            dict_merge(node_attrs, attrs, overwrite=True)



import networkx
from typing import Iterable, Set, Dict
from collections import defaultdict
from ..node import NodeIdType
from ..inittask import get_task_class


def graph_is_cyclic(graph: networkx.DiGraph) -> bool:
    return not networkx.is_directed_acyclic_graph(graph)


def graph_has_conditional_links(graph: networkx.DiGraph) -> bool:
    for attrs in graph.edges.values():
        if attrs.get("conditions") or attrs.get("on_error"):
            return True
    return False


def node_successors(
    graph: networkx.DiGraph, node_id: NodeIdType, **include_filter
) -> Iterable[NodeIdType]:
    if include_filter:
        yield from iter_downstream_nodes(
            graph, node_id, recursive=False, **include_filter
        )
    else:
        yield from graph.successors(node_id)


def node_descendants(
    graph: networkx.DiGraph, node_id: NodeIdType, **include_filter
) -> Iterable[NodeIdType]:
    yield from iter_downstream_nodes(graph, node_id, recursive=True, **include_filter)


def node_predecessors(
    graph: networkx.DiGraph, node_id: NodeIdType, **include_filter
) -> Iterable[NodeIdType]:
    if include_filter:
        yield from iter_upstream_nodes(
            graph, node_id, recursive=False, **include_filter
        )
    else:
        yield from graph.predecessors(node_id)


def node_ancestors(
    graph: networkx.DiGraph, node_id: NodeIdType, **include_filter
) -> Iterable[NodeIdType]:
    yield from iter_upstream_nodes(graph, node_id, recursive=True, **include_filter)


def iterator_has_items(iterator):
    try:
        next(iterator)
        return True
    except StopIteration:
        return False


def node_has_successors(graph: networkx.DiGraph, node_id: NodeIdType, **include_filter):
    return iterator_has_items(node_successors(graph, node_id, **include_filter))


def node_has_descendants(
    graph: networkx.DiGraph, node_id: NodeIdType, **include_filter
):
    return iterator_has_items(node_descendants(graph, node_id, **include_filter))


def node_has_predecessors(
    graph: networkx.DiGraph, node_id: NodeIdType, **include_filter
):
    return iterator_has_items(node_predecessors(graph, node_id, **include_filter))


def node_has_ancestors(graph: networkx.DiGraph, node_id: NodeIdType, **include_filter):
    return iterator_has_items(node_ancestors(graph, node_id, **include_filter))


def iter_downstream_nodes(
    graph: networkx.DiGraph, node_id: NodeIdType, **kw
) -> Iterable[NodeIdType]:
    yield from _iter_nodes(graph, node_id, upstream=False, **kw)


def iter_upstream_nodes(
    graph: networkx.DiGraph, node_id: NodeIdType, **kw
) -> Iterable[NodeIdType]:
    yield from _iter_nodes(graph, node_id, upstream=True, **kw)


def _iter_nodes(
    graph: networkx.DiGraph,
    node_id: NodeIdType,
    upstream=False,
    recursive=False,
    _visited=None,
    **include_filter,
) -> Iterable[NodeIdType]:
    """Recursion is not stopped by the node or link filters.
    Recursion is stopped by either not having any successors/predecessors
    or encountering a node that has been visited already.
    The original node on which we start iterating is never yielded.
    """
    if recursive:
        if _visited is None:
            _visited = set()
        elif node_id in _visited:
            return
        _visited.add(node_id)
    if upstream:
        iter_next_nodes = graph.predecessors
    else:
        iter_next_nodes = graph.successors
    if not include_filter.get("parsed"):
        include_filter = {f"_{k}": v for k, v in include_filter.items()}
        include_filter["parsed"] = True
    for next_id in iter_next_nodes(node_id):
        node_is_included = _filter_node(graph, next_id, **include_filter)
        if upstream:
            link_is_included = _filter_link(graph, next_id, node_id, **include_filter)
        else:
            link_is_included = _filter_link(graph, node_id, next_id, **include_filter)
        if node_is_included and link_is_included:
            yield next_id
        if recursive:
            yield from _iter_nodes(
                graph,
                next_id,
                upstream=upstream,
                recursive=True,
                _visited=_visited,
                **include_filter,
            )


def _filter_node(
    graph: networkx.DiGraph,
    node_id: NodeIdType,
    _node_filter=None,
    _node_has_predecessors=None,
    _node_has_successors=None,
    _node_has_error_handlers=None,
    **_,
) -> bool:
    """Filters are combined with the logical AND"""
    if callable(_node_filter):
        if not _node_filter(node_id):
            return False
    if _node_has_predecessors is not None:
        if node_has_predecessors(graph, node_id) != _node_has_predecessors:
            return False
    if _node_has_successors is not None:
        if node_has_successors(graph, node_id) != _node_has_successors:
            return False
    if _node_has_error_handlers is not None:
        if node_has_error_handlers(graph, node_id) != _node_has_error_handlers:
            return False
    return True


def _filter_link(
    graph: networkx.DiGraph,
    source_id: NodeIdType,
    target_id: NodeIdType,
    _link_filter=None,
    _link_has_on_error=None,
    _link_has_conditions=None,
    _link_is_conditional=None,
    _link_has_required=None,
    **_,
) -> bool:
    """Filters are combined with the logical AND"""
    if callable(_link_filter):
        if not _link_filter(source_id, target_id):
            return False
    if _link_has_on_error is not None:
        if link_has_on_error(graph, source_id, target_id) != _link_has_on_error:
            return False
    if _link_has_conditions is not None:
        if link_has_conditions(graph, source_id, target_id) != _link_has_conditions:
            return False
    if _link_is_conditional is not None:
        if link_is_conditional(graph, source_id, target_id) != _link_is_conditional:
            return False
    if _link_has_required is not None:
        if link_has_required(graph, source_id, target_id) != _link_has_required:
            return False
    return True


def link_has_conditions(
    graph: networkx.DiGraph, source_id: NodeIdType, target_id: NodeIdType
) -> bool:
    link_attrs = graph[source_id][target_id]
    return bool(link_attrs.get("conditions", False))


def link_has_on_error(
    graph: networkx.DiGraph, source_id: NodeIdType, target_id: NodeIdType
) -> bool:
    link_attrs = graph[source_id][target_id]
    return bool(link_attrs.get("on_error", False))


def link_has_required(
    graph: networkx.DiGraph, source_id: NodeIdType, target_id: NodeIdType
) -> bool:
    link_attrs = graph[source_id][target_id]
    return bool(link_attrs.get("required", False))


def link_is_conditional(
    graph: networkx.DiGraph, source_id: NodeIdType, target_id: NodeIdType
) -> bool:
    link_attrs = graph[source_id][target_id]
    return bool(
        link_attrs.get("on_error", False) or link_attrs.get("conditions", False)
    )


def link_is_required(
    graph: networkx.DiGraph, source_id: NodeIdType, target_id: NodeIdType
) -> bool:
    if link_has_required(graph, source_id, target_id):
        return True
    if link_is_conditional(graph, source_id, target_id):
        return False
    return node_is_required(graph, source_id)


def node_is_required(graph: networkx.DiGraph, node_id: NodeIdType) -> bool:
    not_required = node_has_ancestors(
        graph, node_id, link_has_required=False, link_is_conditional=True
    )
    not_required |= node_has_ancestors(graph, node_id, node_has_error_handlers=True)
    return not not_required


def node_has_error_handlers(graph: networkx.DiGraph, node_id: NodeIdType):
    return node_has_successors(graph, node_id, link_has_on_error=True)


def required_predecessors(
    graph: networkx.DiGraph, target_id: NodeIdType
) -> Iterable[NodeIdType]:
    for source_id in node_predecessors(graph, target_id):
        if link_is_required(graph, source_id, target_id):
            yield source_id


def has_required_predecessors(graph: networkx.DiGraph, node_id: NodeIdType) -> bool:
    return iterator_has_items(required_predecessors(graph, node_id))


def has_required_static_inputs(graph: networkx.DiGraph, node_id: NodeIdType) -> bool:
    """Returns True when the default inputs cover all required inputs."""
    node_attrs = graph.nodes[node_id]
    inputs_complete = node_attrs.get("inputs_complete", None)
    if isinstance(inputs_complete, bool):
        # method and script tasks always have an empty `required_input_names`
        # although they may have required input. This keyword is used the
        # manually indicate that all required inputs are statically provided.
        return inputs_complete
    taskclass = get_task_class(node_id, node_attrs)
    static_inputs = {d["name"] for d in node_attrs.get("default_inputs", list())}
    return not (set(taskclass.required_input_names()) - static_inputs)


def node_condition_values(
    graph: networkx.DiGraph, source_id: NodeIdType
) -> Dict[str, set]:
    condition_values = defaultdict(set)
    for target_id in node_successors(graph, source_id, link_has_conditions=True):
        for condition in graph[source_id][target_id]["conditions"]:
            varname = condition["source_output"]
            value = condition["value"]
            condition_values[varname].add(value)
    return condition_values


def node_has_noncovered_conditions(
    graph: networkx.DiGraph, source_id: NodeIdType
) -> bool:
    conditions_else_value = graph.nodes[source_id].get("conditions_else_value", None)
    complements = {
        True: {False, conditions_else_value},
        False: {True, conditions_else_value},
    }
    condition_values = node_condition_values(graph, source_id)
    for values in condition_values.values():
        for value in values:
            cvalue = complements.get(value, conditions_else_value)
            if cvalue not in values:
                return True
    return False


def start_nodes(graph: networkx.DiGraph) -> Set[NodeIdType]:
    """Nodes from which the graph execution starts"""
    nodes = set(
        node_id for node_id in graph.nodes if not node_has_predecessors(graph, node_id)
    )
    if nodes:
        return nodes
    return set(
        node_id
        for node_id in graph.nodes
        if has_required_static_inputs(graph, node_id)
        and not has_required_predecessors(graph, node_id)
    )


def end_nodes(graph: networkx.DiGraph) -> Set[NodeIdType]:
    """Nodes at which an graph execution thread may end and
    which result need to be recorded.
    """
    nodes = set(
        node_id for node_id in graph.nodes if node_is_pure_end_node(graph, node_id)
    )
    if nodes:
        return nodes
    return set(node_id for node_id in graph.nodes if node_is_end_node(graph, node_id))


def node_is_pure_end_node(graph: networkx.DiGraph, node_id: NodeIdType) -> bool:
    """Node without successors or only error handlers"""
    return not node_has_successors(graph, node_id, link_has_on_error=False)


def node_is_end_node(graph: networkx.DiGraph, node_id: NodeIdType) -> bool:
    """A pure end node or a node with uncovered conditions"""
    if node_is_pure_end_node(graph, node_id):
        return True
    if node_has_noncovered_conditions(graph, node_id):
        return True
    return False


def topological_sort(graph: networkx.DiGraph) -> Iterable[NodeIdType]:
    """Sort node names for sequential instantiation+execution of DAGs"""
    if graph_is_cyclic(graph):
        raise RuntimeError("Sorting nodes is not possible for cyclic graphs")
    yield from networkx.topological_sort(graph)


def node_pure_descendants(
    graph: networkx.DiGraph, node_id: NodeIdType, include_node: bool = False
) -> Iterable[NodeIdType]:
    """Yields all descendants which do not depend on anything else than `node_id`"""
    if include_node:
        yield node_id
    nodes = {node_id}
    iter_successors = {node_id}
    while iter_successors:
        new_iter_successors = set()
        for node_id in iter_successors:
            for target_id in graph.successors(node_id):
                if target_id in nodes:
                    continue  # loopback
                predecessors = set(graph.predecessors(target_id))
                if predecessors - nodes:
                    continue  # depends on a node outside the branch
                yield target_id
                new_iter_successors.add(target_id)
        iter_successors = new_iter_successors



from .task import Task
from ewoksutils.import_utils import import_method

METHOD_ARGUMENT = "_method"
PPF_DICT_ARGUMENT = "_ppfdict"


class PpfMethodExecutorTask(
    Task,
    input_names=[METHOD_ARGUMENT],
    optional_input_names=[PPF_DICT_ARGUMENT],
    output_names=[PPF_DICT_ARGUMENT],
):
    """Ppf workflows pass one dictionary around between tasks and this dictionary
    gets updates by each task. This dictionary is unpacked into the unexpected
    arguments and passed to the method.
    """

    METHOD_ARGUMENT = METHOD_ARGUMENT
    PPF_DICT_ARGUMENT = PPF_DICT_ARGUMENT

    def run(self):
        method_kwargs = self.get_input_values()
        fullname = method_kwargs.pop(self.METHOD_ARGUMENT)
        method = import_method(fullname)
        ppfdict = method_kwargs.pop(self.PPF_DICT_ARGUMENT, None)
        if ppfdict:
            method_kwargs.update(ppfdict)

        result = method(**method_kwargs)

        method_kwargs.update(result)
        self.outputs._ppfdict = method_kwargs


class PpfPortTask(
    Task, optional_input_names=[PPF_DICT_ARGUMENT], output_names=[PPF_DICT_ARGUMENT]
):
    """A ppfmethod which represents the identity mapping"""

    PPF_DICT_ARGUMENT = PPF_DICT_ARGUMENT

    def run(self):
        method_kwargs = self.get_input_values()
        ppfdict = method_kwargs.pop(self.PPF_DICT_ARGUMENT, None)
        if ppfdict:
            method_kwargs.update(ppfdict)
        self.outputs._ppfdict = method_kwargs



import logging
from ewoksutils.event_utils import FIELD_TYPES

__all__ = ["is_ewoks_event_handler", "EwoksEventHandlerMixIn", "EwoksEventHandler"]


def is_ewoks_event_handler(handler):
    return isinstance(handler, EwoksEventHandlerMixIn)


class EwoksEventHandlerMixIn:
    BLOCKING = False
    FIELD_TYPES = FIELD_TYPES


class EwoksEventHandler(EwoksEventHandlerMixIn, logging.Handler):
    """Base class for handling ewoks events on the publishing side (implement the `emit` method)."""

    pass



from typing import Any, List, Mapping
from pathlib import Path

import h5py
from silx.io.dictdump import dicttonx, nxtodict
from silx.io import h5py_utils
from silx.utils.proxy import docstring

from .file import FileProxy
from . import atomic


# @h5py_utils.retry(retry_period=1)
def h5_item_exists(path, item):
    try:
        with h5py_utils.File(path) as f:
            return item in f
    except Exception:
        return False


class NexusProxy(FileProxy):
    SCHEME = "nexus"
    EXTENSIONS = [".nx", "nxs", ".h5", ".hdf5", ".nexus"]
    ALLOW_PATH_IN_FILE = True

    @docstring(FileProxy)
    def exists(self) -> bool:
        if not super().exists():
            return False
        return h5_item_exists(self.path, self.path_in_file)

    def _dump(self, path: Path, data: Any, update_mode: str = "add", **_) -> None:
        if isinstance(data, Mapping):
            h5group = self.path_in_file
        else:
            h5group = self.path_in_file_parent
            h5name = self.path_in_file_name
            if h5name:
                data = {h5name: data}
            elif not isinstance(data, Mapping):
                raise TypeError("'data' must be a dictionary")
        with atomic.atomic_write_hdf5(path, h5group) as (h5file, h5group):
            dicttonx(
                treedict=data,
                h5file=h5file,
                h5path=h5group,
                update_mode=update_mode,
                add_nx_class=True,
            )
            self._set_nx_classes(h5file, self.path_in_file_parts())

    @staticmethod
    def _set_nx_classes(h5file, parts: List[str]):
        h5file.attrs["NX_class"] = "NXroot"
        h5group = h5file
        for i, name in enumerate(parts):
            h5group = h5group[name]
            if not isinstance(h5group, h5py.Group):
                break
            if i == 0:
                NX_class = "NXentry"
            elif i == 1:
                NX_class = "NXprocess"
            else:
                NX_class = "NXcollection"
            h5group.attrs["NX_class"] = NX_class

    def _load(self, path: Path, **kw) -> Any:
        h5group = self.path_in_file_parent
        h5name = self.path_in_file_name
        adict = nxtodict(h5file=str(path), path=h5group, **kw)
        if h5name:
            return adict[h5name]
        return adict



from . import graph


@graph
def cyclic1():
    task = "ewokscore.tests.examples.tasks.condsumtask.CondSumTask"
    nodes = [
        {
            "id": "task1",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task2",
            "default_inputs": [{"name": "b", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task3",
            "default_inputs": [{"name": "b", "value": 3}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task4",
            "default_inputs": [{"name": "b", "value": -1}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task5",
            "default_inputs": [{"name": "b", "value": -1}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task6",
            "default_inputs": [{"name": "b", "value": 0}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task7",
            "default_inputs": [{"name": "b", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
        },
    ]

    links = [
        {
            "source": "task1",
            "target": "task2",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task3",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task3",
            "target": "task4",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task4",
            "target": "task2",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
            "conditions": [{"source_output": "too_small", "value": True}],
        },
        {
            "source": "task4",
            "target": "task5",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
            "conditions": [{"source_output": "too_small", "value": False}],
        },
        {
            "source": "task5",
            "target": "task6",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task6",
            "target": "task2",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
            "conditions": [{"source_output": "too_small", "value": True}],
        },
        {
            "source": "task6",
            "target": "task7",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
            "conditions": [{"source_output": "too_small", "value": False}],
        },
    ]

    expected = {"task7": {"result": 12, "too_small": False}}

    graph = {
        "links": links,
        "nodes": nodes,
    }

    return graph, expected



import os
from pathlib import Path
from ewokscore.persistence.uri import parse_uri
from ewokscore.persistence.uri import path_from_uri
from ewokscore.persistence.uri import parse_query


def test_file_uri():
    nonpath = str(Path("file.h5"))
    relpath = str(Path("relpath") / "file.h5")
    abspath = str(Path(os.path.sep) / "abspath" / "file.h5")
    assert not os.path.isabs(nonpath)
    assert not os.path.isabs(relpath)
    assert os.path.isabs(abspath)

    uri = nonpath
    assert parse_uri(uri).geturl() == "file:///file.h5"
    assert str(path_from_uri(parse_uri(uri))) == uri
    assert parse_query(parse_uri(uri)) == dict()

    uri = relpath
    assert parse_uri(uri).geturl() == "file:///relpath/file.h5"
    assert str(path_from_uri(parse_uri(uri))) == uri
    assert parse_query(parse_uri(uri)) == dict()

    uri = abspath
    assert parse_uri(uri).geturl() == "file:///abspath/file.h5"
    assert str(path_from_uri(parse_uri(uri))) == uri
    assert parse_query(parse_uri(uri)) == dict()

    uri = nonpath + "::/entry"
    assert parse_uri(uri).geturl() == "file:///file.h5?path=/entry"
    assert str(path_from_uri(parse_uri(uri))) == nonpath
    assert parse_query(parse_uri(uri)) == {"path": "/entry"}

    uri = relpath + "::/entry"
    assert parse_uri(uri).geturl() == "file:///relpath/file.h5?path=/entry"
    assert str(path_from_uri(parse_uri(uri))) == relpath
    assert parse_query(parse_uri(uri)) == {"path": "/entry"}

    uri = abspath + "::/entry"
    assert parse_uri(uri).geturl() == "file:///abspath/file.h5?path=/entry"
    assert str(path_from_uri(parse_uri(uri))) == abspath
    assert parse_query(parse_uri(uri)) == {"path": "/entry"}

    uri = abspath + "::/entry?name=abc"
    assert parse_uri(uri).geturl() == "file:///abspath/file.h5?path=/entry&name=abc"
    assert str(path_from_uri(parse_uri(uri))) == abspath
    assert parse_query(parse_uri(uri)) == {"path": "/entry", "name": "abc"}

    uri = abspath + "::/entry?path=xyz&name=abc"
    assert parse_uri(uri).geturl() == "file:///abspath/file.h5?path=/entry/xyz&name=abc"
    assert str(path_from_uri(parse_uri(uri))) == abspath
    assert parse_query(parse_uri(uri)) == {"path": "/entry/xyz", "name": "abc"}



import pytest
from ewokscore import missing_data
from ewokscore.variable import MutableVariableContainer
from ewokscore.variable import VariableContainerNamespace
from ewokscore.variable import ReadOnlyVariableContainerNamespace
from ewokscore.variable import VariableContainerMissingNamespace
from ewokscore.variable import MissingVariableError
from ewokscore.variable import ReadOnlyVariableError


def test_namespace():
    variables = MutableVariableContainer(
        value={"a": None, 0: 1, "c": missing_data.MISSING_DATA}
    )
    namespace = VariableContainerNamespace(variables)
    assert namespace.a is None
    assert namespace["a"] is None
    assert namespace[0] == 1
    assert namespace.c is missing_data.MISSING_DATA
    assert namespace["c"] is missing_data.MISSING_DATA
    with pytest.raises(MissingVariableError):
        namespace.wrong
    with pytest.raises(MissingVariableError):
        namespace["wrong"]
    namespace.a = 10
    assert namespace.a == 10
    namespace["a"] = 20
    assert namespace["a"] == 20


def test_readonly_namespace():
    variables = MutableVariableContainer(
        value={"a": None, 0: 1, "c": missing_data.MISSING_DATA}
    )
    namespace = ReadOnlyVariableContainerNamespace(variables)
    assert namespace.a is None
    assert namespace["a"] is None
    assert namespace[0] == 1
    assert namespace.c is missing_data.MISSING_DATA
    assert namespace["c"] is missing_data.MISSING_DATA
    with pytest.raises(MissingVariableError):
        namespace.wrong
    with pytest.raises(MissingVariableError):
        namespace["wrong"]
    with pytest.raises(ReadOnlyVariableError):
        namespace.a = 10
    with pytest.raises(ReadOnlyVariableError):
        namespace["a"] = 10


def test_missing_namespace():
    variables = MutableVariableContainer(
        value={"a": None, 0: 1, "c": missing_data.MISSING_DATA}
    )
    namespace = VariableContainerMissingNamespace(variables)
    assert namespace.a is False
    assert namespace["a"] is False
    assert namespace[0] is False
    assert namespace.c is True
    assert namespace["c"] is True
    with pytest.raises(MissingVariableError):
        namespace.wrong
    with pytest.raises(MissingVariableError):
        namespace["wrong"]
    with pytest.raises(ReadOnlyVariableError):
        namespace.a = True
    with pytest.raises(ReadOnlyVariableError):
        namespace["a"] = True



import json
from typing import Tuple, Any


def parse_value(value: str) -> Any:
    try:
        return json.loads(value)
    except Exception:
        return value


_NODE_ATTR_MAP = {"id": "id", "label": "label", "taskid": "task_identifier"}


def parse_parameter(input_item: str, node_attr: str, all: bool) -> dict:
    """The format of `input_item` is `"[NODE]:name=value"`"""
    node_and_name, _, value = input_item.partition("=")
    a, sep, b = node_and_name.partition(":")
    if sep:
        node = a
        var_name = b
    else:
        node = None
        var_name = a
    var_value = parse_value(value)
    if node is None:
        return {"all": all, "name": var_name, "value": var_value}
    return {
        _NODE_ATTR_MAP[node_attr]: node,
        "name": var_name,
        "value": var_value,
    }


def parse_option(option: str) -> Tuple[str, Any]:
    option, _, value = option.partition("=")
    return option, parse_value(value)


def parse_workflow(args):
    if args.test:
        from ewokscore.tests.examples.graphs import graph_names, get_graph

        graphs = list(graph_names())
        if args.workflow not in graphs:
            raise RuntimeError(f"Test graph '{args.workflow}' does not exist: {graphs}")

        graph, _ = get_graph(args.workflow)
    else:
        graph = args.workflow
    return graph



import json
import yaml
import pytest
from ewokscore import execute_graph
from ewokscore.graph import load_graph


def subsubsubgraph():
    return {
        "graph": {
            "id": "subsubsubgraph",
            "input_nodes": [{"id": "in", "node": "task1"}, {"id": "notconnected"}],
            "output_nodes": [{"id": "out", "node": "task2"}, {"id": "notconnected"}],
        },
        "nodes": [
            {
                "id": "task1",
                "task_type": "method",
                "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.add",
            },
            {
                "id": "task2",
                "task_type": "method",
                "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.add",
            },
        ],
        "links": [
            {
                "source": "task1",
                "target": "task2",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            },
        ],
    }


def subsubgraph(_subsubsubgraph):
    return {
        "graph": {
            "id": "subsubgraph",
            "input_nodes": [{"id": "in", "node": "task1"}, {"id": "notconnected"}],
            "output_nodes": [
                {"id": "out", "node": "subsubsubgraph", "sub_node": "out"},
                {"id": "notconnected"},
            ],
        },
        "nodes": [
            {
                "id": "task1",
                "task_type": "method",
                "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.add",
            },
            {
                "id": "task2",
                "task_type": "method",
                "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.add",
            },
            {
                "id": "subsubsubgraph",
                "task_type": "graph",
                "task_identifier": _subsubsubgraph,
            },
        ],
        "links": [
            {
                "source": "task1",
                "target": "task2",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            },
            {
                "source": "task2",
                "target": "subsubsubgraph",
                "sub_target": "in",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            },
        ],
    }


def subgraph(_subsubgraph):
    return {
        "graph": {
            "id": "subgraph",
            "input_nodes": [{"id": "in", "node": "task1"}],
            "output_nodes": [{"id": "out", "node": "subsubgraph", "sub_node": "out"}],
        },
        "nodes": [
            {
                "id": "task1",
                "task_type": "method",
                "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.add",
            },
            {
                "id": "task2",
                "task_type": "method",
                "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.add",
            },
            {
                "id": "subsubgraph",
                "task_type": "graph",
                "task_identifier": _subsubgraph,
            },
        ],
        "links": [
            {
                "source": "task1",
                "target": "task2",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            },
            {
                "source": "task2",
                "target": "subsubgraph",
                "sub_target": "in",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            },
        ],
    }


def graph(_subgraph):
    return {
        "graph": {"id": "graph"},
        "nodes": [
            {"id": "subgraph1", "task_type": "graph", "task_identifier": _subgraph},
            {"id": "subgraph2", "task_type": "graph", "task_identifier": _subgraph},
            {
                "id": "append",
                "task_type": "method",
                "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.append",
            },
        ],
        "links": [
            {
                "source": "subgraph1",
                "sub_source": "out",
                "target": "subgraph2",
                "sub_target": "in",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            },
            # Link all nodes from "subgraph1" to "append"
            {
                "source": "subgraph1",
                "sub_source": "task1",
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            },
            {
                "source": "subgraph1",
                "sub_source": "task2",
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 1}],
            },
            {
                "source": "subgraph1",
                "sub_source": ("subsubgraph", "task1"),
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 2}],
            },
            {
                "source": "subgraph1",
                "sub_source": ("subsubgraph", "task2"),
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 3}],
            },
            {
                "source": "subgraph1",
                "sub_source": ("subsubgraph", ("subsubsubgraph", "task1")),
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 4}],
            },
            {
                "source": "subgraph1",
                "sub_source": ("subsubgraph", ("subsubsubgraph", "task2")),
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 5}],
            },
            # Link all nodes from "subgraph2" to "append"
            {
                "source": "subgraph2",
                "sub_source": "task1",
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 6}],
            },
            {
                "source": "subgraph2",
                "sub_source": "task2",
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 7}],
            },
            {
                "source": "subgraph2",
                "sub_source": ("subsubgraph", "task1"),
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 8}],
            },
            {
                "source": "subgraph2",
                "sub_source": ("subsubgraph", "task2"),
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 9}],
            },
            {
                "source": "subgraph2",
                "sub_source": ("subsubgraph", ("subsubsubgraph", "task1")),
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 10}],
            },
            {
                "source": "subgraph2",
                "sub_source": ("subsubgraph", ("subsubsubgraph", "task2")),
                "target": "append",
                "data_mapping": [{"source_output": "return_value", "target_input": 11}],
            },
        ],
    }


def savegraph(graph, tmpdir, name, representation="json"):
    if representation == "json":
        filename = name + ".json"
        with open(tmpdir / filename, mode="w") as f:
            json.dump(graph, f, indent=2)
    elif representation == "yaml":
        filename = name + ".yml"
        with open(tmpdir / filename, mode="w") as f:
            yaml.dump(graph, f)
    else:
        raise ValueError(representation)
    return filename


def serialized_graph(tmpdir, representation="json") -> str:
    subname = savegraph(
        subsubsubgraph(), tmpdir, "subsubsubgraph", representation=representation
    )
    subname = savegraph(
        subsubgraph(subname), tmpdir, "subsubgraph", representation=representation
    )
    subname = savegraph(
        subgraph(subname), tmpdir, "subgraph", representation=representation
    )
    return savegraph(graph(subname), tmpdir, "graph", representation=representation)


def nonserialized_graph() -> dict:
    return graph(subgraph(subsubgraph(subsubsubgraph())))


@pytest.mark.parametrize(
    "representation", (None, "json", "json_dict", "json_string", "yaml")
)
def test_sub_graph_serialize(representation, tmpdir):
    ewoksgraph = load_graph(nonserialized_graph())
    if representation == "yaml":
        destination = str(tmpdir / "file.yml")
    elif representation == "json":
        destination = str(tmpdir / "file.json")
    else:
        destination = None
    inmemorydump = ewoksgraph.dump(destination, representation=representation)

    if destination:
        source = destination
    else:
        source = inmemorydump
    ewoksgraph2 = load_graph(source, representation=representation)

    assert ewoksgraph == ewoksgraph2


@pytest.mark.parametrize("representation", ("json", "yaml"))
def test_sub_graph_execute(representation, tmpdir):
    g = serialized_graph(tmpdir, representation=representation)
    ewoksgraph = load_graph(g, root_dir=str(tmpdir))

    tasks = execute_graph(ewoksgraph, output_tasks=True)

    assert len(tasks) == 13

    task = tasks[("subgraph1", "task1")]
    assert task.outputs.return_value == 1
    task = tasks[("subgraph1", "task2")]
    assert task.outputs.return_value == 2
    task = tasks[("subgraph1", ("subsubgraph", "task1"))]
    assert task.outputs.return_value == 3
    task = tasks[("subgraph1", ("subsubgraph", "task2"))]
    assert task.outputs.return_value == 4
    task = tasks[("subgraph1", ("subsubgraph", ("subsubsubgraph", "task1")))]
    assert task.outputs.return_value == 5
    task = tasks[("subgraph1", ("subsubgraph", ("subsubsubgraph", "task2")))]
    assert task.outputs.return_value == 6

    task = tasks[("subgraph2", "task1")]
    assert task.outputs.return_value == 7
    task = tasks[("subgraph2", "task2")]
    assert task.outputs.return_value == 8
    task = tasks[("subgraph2", ("subsubgraph", "task1"))]
    assert task.outputs.return_value == 9
    task = tasks[("subgraph2", ("subsubgraph", "task2"))]
    assert task.outputs.return_value == 10
    task = tasks[("subgraph2", ("subsubgraph", ("subsubsubgraph", "task1")))]
    assert task.outputs.return_value == 11
    task = tasks[("subgraph2", ("subsubgraph", ("subsubsubgraph", "task2")))]
    assert task.outputs.return_value == 12

    task = tasks["append"]
    assert task.outputs.return_value == tuple(range(1, 13))



import networkx
from pprint import pprint
from ewokscore.graph import load_graph


def show_graph(graph, stdout=True, plot=True, show=True):
    taskgraph = load_graph(graph)
    if stdout:
        pprint(taskgraph.dump())
    if plot:
        networkx.draw(taskgraph.graph, with_labels=True, font_size=10)
        if show:
            import matplotlib.pyplot as plt

            plt.show()



from typing import Optional, List
from ewoksutils import import_utils


class Registered:
    _SUBCLASS_REGISTRY = None

    def __init_subclass__(cls, register=True, registry_name=None, **kwargs):
        super().__init_subclass__(**kwargs)

        # Ensures that not all subclasses share the same registry
        if cls._SUBCLASS_REGISTRY is None:
            cls._SUBCLASS_REGISTRY = dict()

        if not register:
            cls.__REGISTRY_NAME = None
            return

        # Register the subclass
        if not registry_name:
            registry_name = import_utils.qualname(cls)
        ecls = cls._SUBCLASS_REGISTRY.get(registry_name)
        if ecls is not None:
            if import_utils.qualname(cls) == import_utils.qualname(ecls):
                return
            raise NotImplementedError(
                f"Registry name {registry_name} is already taken by {repr(ecls)}"
            )
        cls.__REGISTRY_NAME = registry_name
        cls._SUBCLASS_REGISTRY[registry_name] = cls

    @classmethod
    def class_registry_name(cls) -> Optional[str]:
        return cls.__REGISTRY_NAME

    @classmethod
    def get_subclass_names(cls) -> List[str]:
        return list(cls._SUBCLASS_REGISTRY.keys())

    @classmethod
    def get_subclasses(cls):
        return list(cls._SUBCLASS_REGISTRY.values())

    @classmethod
    def get_subclass(cls, registry_name, _second_attempt=False):
        """Retrieving a derived class"""
        subclass = cls._SUBCLASS_REGISTRY.get(registry_name)
        if subclass is None:
            candidates = [
                name
                for name in cls._SUBCLASS_REGISTRY
                if name.endswith("." + registry_name)
            ]
            if len(candidates) == 1:
                subclass = cls._SUBCLASS_REGISTRY.get(candidates[0])
        if subclass is None:
            if _second_attempt:
                lst = cls.get_subclass_names()
                raise RuntimeError(
                    f"Class {repr(registry_name)} is not imported. Imported classes are {repr(lst)}"
                )
            else:
                import_utils.import_qualname(registry_name)
                subclass = cls.get_subclass(registry_name, _second_attempt=True)
        return subclass



from . import graph
from ewokscore import missing_data


@graph
def acyclic3():
    task = "ewokscore.tests.examples.tasks.sumtask.SumTask"
    nodes = [
        {
            "id": "task1",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task2",
            "default_inputs": [{"name": "a", "value": 2}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task3",
            "default_inputs": [{"name": "b", "value": 3}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task4",
            "default_inputs": [{"name": "b", "value": 4}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task5",
            "default_inputs": [{"name": "b", "value": 5}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task6",
            "default_inputs": [{"name": "b", "value": 6}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task7",
            "task_type": "class",
            "task_identifier": "ewokscore.tests.examples.tasks.nooutputtask.NoOutputTask",
        },
        {
            "id": "task8",
            "task_type": "class",
            "task_identifier": "ewokscore.tests.examples.tasks.nooutputtask.NoOutputTask",
        },
    ]

    links = [
        {
            "source": "task1",
            "target": "task3",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task4",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task3",
            "target": "task5",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task4",
            "target": "task5",
            "data_mapping": [{"target_input": "b", "source_output": "result"}],
        },
        {
            "source": "task5",
            "target": "task6",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task6",
            "target": "task7",
            "map_all_data": True,
        },
        {
            "source": "task7",
            "target": "task8",
        },
    ]

    graph = {
        "links": links,
        "nodes": nodes,
    }

    expected_results = {
        "task1": {"result": 1},
        "task2": {"result": 2},
        "task3": {"result": 4},
        "task4": {"result": 6},
        "task5": {"result": 10},
        "task6": {"result": 16},
        "task7": missing_data.MISSING_DATA,
        "task8": missing_data.MISSING_DATA,
    }

    return graph, expected_results



import pytest
from jupyter_client.kernelspec import KernelSpecManager
from ipykernel.kernelspec import install as install_kernel


@pytest.fixture
def varinfo(tmpdir):
    yield {"root_uri": str(tmpdir)}


@pytest.fixture(scope="session")
def testkernel():
    m = KernelSpecManager()
    kernel_name = "pytest_kernel"
    install_kernel(kernel_name=kernel_name, user=True)
    yield kernel_name
    m.remove_kernel_spec(kernel_name)
